/*
    ChemCon - molecular mechanics and molecular graphics
    Copyright (C) 1998-2002  Alexei Nikitin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Painter_3D.h"

#include <math.h>

#include "Image_3D.h"
#include "Z_buffer.h"
#include "Material.h"
#include "ViewObj.h"


namespace
{
const int MAX_RAD = 150;
const int shift = 2;        //fix
}//namespace

Painter_3D::
Painter_3D( Image_3D & image_3D )
:
    Painter_2D( image_3D.image() ),
    image_3D_( image_3D ),
    z_buf_( & image_3D.z_buffer() ) //fix
{
}

void Painter_3D::
line( int XStart, int YStart, int XEnd, int YEnd )
  {
    int X, Y;
    int Temp, AdjUp, AdjDown, ErrorTerm, XAdvance, XDelta, YDelta;
    int WholeStep, InitialPixelCount, FinalPixelCount, i, RunLength;
///    char far *ScreenPtr;

    /* We'll always draw top to bottom, to reduce the number of cases we have to
    handle, and to make lines between the same endpoints draw the same pixels */
    if (YStart > YEnd) {
      Temp = YStart;
      YStart = YEnd;
      YEnd = Temp;
      Temp = XStart;
      XStart = XEnd;
      XEnd = Temp;
    }

    X = XStart;
    Y = YStart;
    // Point to the bitmap address first pixel to draw 
    //   ScreenPtr = MK_FP(SCREEN_SEGMENT, YStart * SCREEN_WIDTH + XStart);

    // Figure out whether we're going left or right, and how far we're
    //   going horizontally 
    if ((XDelta = XEnd - XStart) < 0)
    {
      XAdvance = -1;
      XDelta = -XDelta;
    }
    else
    {
      XAdvance = 1;
    }
    // Figure out how far we're going vertically 
    YDelta = YEnd - YStart;

    // Special-case horizontal, vertical, and diagonal lines, for speed
    // and to avoid nasty boundary conditions and division by 0 
    /*if (XDelta == 0)
    {
      // Vertical line 
      for (i=0; i<=YDelta; i++)
      {
        /// *ScreenPtr = Color;
        ///ScreenPtr += SCREEN_WIDTH;
        point( X, Y++ );
      }
      return;
    }//*/
    if (YDelta == 0)
    {
      // Horizontal line 
      for (i=0; i<=XDelta; i++)
      {
        // *ScreenPtr = Color;
        //ScreenPtr += XAdvance;
        point( X += XAdvance, Y );
      }
      return;
    }//*/
    /*if (XDelta == YDelta)
    {
      // Diagonal line 
      for (i=0; i<=XDelta; i++)
      {
        // *ScreenPtr = Color;
        //ScreenPtr += XAdvance + SCREEN_WIDTH;
        point( X += XAdvance, Y++ );
      }
      return;
    }
    //*/
    // Determine whether the line is X or Y major, and handle accordingly 
    //if (XDelta >= YDelta)
    {
      // X major line 
      // Minimum # of pixels in a run in this line 
      WholeStep = XDelta / YDelta;

      // Error term adjust each time Y steps by 1; used to tell when one
      //   extra pixel should be drawn as part of a run, to account for
      //   fractional steps along the X axis per 1-pixel steps along Y 
      AdjUp = (XDelta % YDelta) * 2;

      // Error term adjust when the error term turns over, used to factor
      //   out the X step made at that time 
      AdjDown = YDelta * 2;

      // Initial error term; reflects an initial step of 0.5 along the Y
      //   axis 
      ErrorTerm = (XDelta % YDelta) - (YDelta * 2);

      // The initial and last runs are partial, because Y advances only 0.5
      //   for these runs, rather than 1. Divide one full run, plus the
      //   initial pixel, between the initial and last runs 
      InitialPixelCount = (WholeStep / 2) + 1;
      FinalPixelCount = InitialPixelCount;

      // If the basic run length is even and there's no fractional
      //   advance, we have one pixel that could go to either the initial
      //   or last partial run, which we'll arbitrarily allocate to the
      //   last run  
      if ((AdjUp == 0) && ((WholeStep & 0x01) == 0))
      {
        InitialPixelCount--;
      }
      // If there're an odd number of pixels per run, we have 1 pixel that can't
      //   be allocated to either the initial or last partial run, so we'll add 0.5
      //   to error term so this pixel will be handled by the normal full-run loop 
       if ((WholeStep & 0x01) != 0)
        {
          ErrorTerm += YDelta;
        }
      // Draw the first, partial run of pixels 
      ////DrawHorizontalRun(&ScreenPtr, XAdvance, InitialPixelCount, Color);
      point( X, Y );
      X += XAdvance * InitialPixelCount;
      Y ++;
      // Draw all full runs
      for (i=0; i<(YDelta-1); i++)
      {
        RunLength = WholeStep;  // run is at least this long 
        // Advance the error term and add an extra pixel if the error
        //   term so indicates 
        if ((ErrorTerm += AdjUp) > 0)
        {
          RunLength++;
          ErrorTerm -= AdjDown;   // reset the error term
        }
        // Draw this scan line's run 
        ////DrawHorizontalRun(&ScreenPtr, XAdvance, RunLength, Color);
        point( X, Y );
        X += XAdvance * RunLength;
        Y ++;
      }
      // Draw the final run of pixels 
      ///DrawHorizontalRun(&ScreenPtr, XAdvance, FinalPixelCount, Color);
      point( X, Y );

      return;
    }
    /*else
    {
    // Y major line

    // Minimum # of pixels in a run in this line
    WholeStep = YDelta / XDelta;

    // Error term adjust each time X steps by 1; used to tell when 1 extra
    //   pixel should be drawn as part of a run, to account for
    //   fractional steps along the Y axis per 1-pixel steps along X
    AdjUp = (YDelta % XDelta) * 2;

    // Error term adjust when the error term turns over, used to factor
    //   out the Y step made at that time
    AdjDown = XDelta * 2;

    // Initial error term; reflects initial step of 0.5 along the X axis
    ErrorTerm = (YDelta % XDelta) - (XDelta * 2);

    // The initial and last runs are partial, because X advances only 0.5
    //   for these runs, rather than 1. Divide one full run, plus the
    //   initial pixel, between the initial and last runs 
    InitialPixelCount = (WholeStep / 2) + 1;
    FinalPixelCount = InitialPixelCount;

    // If the basic run length is even and there's no fractional advance, we
    //   have 1 pixel that could go to either the initial or last partial run,
    //   which we'll arbitrarily allocate to the last run 
    if ((AdjUp == 0) && ((WholeStep & 0x01) == 0))
    {
     InitialPixelCount--;
    }
    // If there are an odd number of pixels per run, we have one pixel
    //   that can't be allocated to either the initial or last partial
    //   run, so we'll add 0.5 to the error term so this pixel will be
    //   handled by the normal full-run loop
    if ((WholeStep & 0x01) != 0)
    {
     ErrorTerm += XDelta;
    }
    // Draw the first, partial run of pixels
    DrawVerticalRun(&ScreenPtr, XAdvance, InitialPixelCount, Color);

    // Draw all full runs
    for (i=0; i<(XDelta-1); i++)
    {
     RunLength = WholeStep;  // run is at least this long 
     // Advance the error term and add an extra pixel if the error
     //   term so indicates 
     if ((ErrorTerm += AdjUp) > 0)
     {
        RunLength++;
        ErrorTerm -= AdjDown;   // reset the error term 
     }
     // Draw this scan line's run 
     DrawVerticalRun(&ScreenPtr, XAdvance, RunLength, Color);
    }
    // Draw the final run of pixels 
    DrawVerticalRun(&ScreenPtr, XAdvance, FinalPixelCount, Color);
    return;
   }//*/
  }
/*
void Painter_3D::
line( int X0, int Y0, int X1, int Y1 )
  {
    unsigned long ErrorAcc, ErrorAdj;
    int DeltaX, DeltaY, XDir, Temp;

    // Make sure the line runs top to bottom
    if(Y0 > Y1) 
      {
        Temp = Y0; Y0 = Y1; Y1 = Temp;
        Temp = X0; X0 = X1; X1 = Temp;
      }

    point( X0, Y0 );              // draw the initial pixel

    if( ( DeltaX = X1 - X0 ) >= 0 ) 
      {
        XDir = 1;
      } 
    else 
      {
        XDir = -1;
        DeltaX = -DeltaX;        // make DeltaX positive
      }

    if( ( DeltaY = Y1 - Y0 ) == 0 )  // done if only one point in the line
      if( DeltaX == 0 ) return;

    ErrorAcc = 0;   

    // Is this an X-major or Y-major line?
    //*/
    /*if( DeltaY > DeltaX ) 
      {
        // Y-major line; calculate 16-bit fixed-point fractional part of a
        // pixel that X advances each time Y advances 1 pixel
        ErrorAdj = ((((unsigned long)DeltaX << 17) / (unsigned long)DeltaY) + 1) >> 1;

        // Draw all pixels between the first and last
        do 
          {
            ErrorAcc += ErrorAdj;        // calculate error for this pixel 
            if (ErrorAcc & ~0xFFFFL) 
              {
                // The error accumulator turned over, so advance the X coord 
                X0 += XDir;
                ErrorAcc &= 0xFFFFL;      // clear integer part of result
              }

            Y0++;                        // Y-major, so always advance Y
            point( X0, Y0 );

          } 
        while( --DeltaY );
        return;
      }
    // It's an X-major line; calculate 16-bit fixed-point fractional part of a
    // pixel that Y advances each time X advances 1 pixel
//    ErrorAdj = ((((unsigned long)DeltaY << 17) / (unsigned long)DeltaX) + 1) >> 1;
    ErrorAdj = (((unsigned long)DeltaY << 16) / (unsigned long)DeltaX);
    int LY = DeltaY / DeltaX;
    // Draw all remaining pixels
    do 
      {
        ErrorAcc += ErrorAdj;          // calculate error for this pixel
        Y0+=LY;
        if( ErrorAcc & ~0xFFFFL ) 
          {
            // The error accumulator turned over, so advance the Y coord
            Y0++;
            ErrorAcc &= 0xFFFFL;      // clear integer part of result
          }
        X0 += XDir;                      // X-major, so always advance X
        point( X0, Y0 );
      } 
    while( --DeltaX );
  }
//*/

void Painter_3D::
Wu_line( int X0, int Y0, int Z0, int X1, int Y1, int Z1 )
  {
    unsigned short  IntensityShift, ErrorAdj, ErrorAcc, Z_Error;
    unsigned short  ErrorAccTemp, Weighting, WeightingComplementMask;
    short           DeltaX, DeltaY, DeltaZ, Temp, XDir, ZDir;
    Byte            Rbg, Gbg, Bbg;
    Byte            R    = red   ( foreground_color_);
    Byte            G    = green ( foreground_color_);
    Byte            B    = blue  ( foreground_color_);
    int             Z;

    Z0 = Z0 << shift;
    Z1 = Z1 << shift;


    // Make sure the line runs top to bottom
    if( (DeltaY = Y1 - Y0) < 0 )
      {
        Temp = (short)Z0; Z0 = Z1; Z1 = (int)Temp;
        Temp = (short)Y0; Y0 = Y1; Y1 = (int)Temp;
        Temp = (short)X0; X0 = X1; X1 = Temp;
        DeltaY = -DeltaY;
      }

    if( (Z = z_buf_->get( X0, Y0 )) < Z0 )
      {
        z_buf_->set( X0, Y0, Z0 );
        point( X0, Y0 );
      }

    if( (DeltaX = X1 - X0) >= 0 ) { XDir = 1;  }
    else
      {
        XDir = -1;
        DeltaX = -DeltaX; // make DeltaX positive
      }

    if( (DeltaZ = Z1 - Z0) >= 0 )  { ZDir = 1; }
    else
      {
        ZDir = -1;
        DeltaZ = -DeltaZ;
      }
    
    if( DeltaX == 0 && DeltaY == 0 )  return;

    if( DeltaY == 0 )
      {
        // Horizontal line
        int WholeStep = DeltaZ / DeltaX;
        int AdjUp = (DeltaZ % DeltaX) * 2;
        int AdjDown = DeltaX * 2;
        int ErrorTerm = (DeltaZ % DeltaX) - (DeltaX * 2);
        if ((WholeStep & 0x01) != 0) ErrorTerm += DeltaX;

        while (DeltaX-- != 0)
          {
            int RunLength = WholeStep; 
            if ((ErrorTerm += AdjUp) > 0)
              {
                RunLength++;
                ErrorTerm -= AdjDown; 
              }
            Z0 += ZDir * RunLength;

            X0 += XDir;

            if( (Z = z_buf_->get( X0, Y0 )) < Z0 )
              {
                z_buf_->set( X0, Y0, Z0 );
                point( X0, Y0 );
              }
          }
        return;
      }

    if(DeltaX == 0)
      {
        // Vertical line
        int WholeStep = DeltaZ / DeltaY;
        int AdjUp = (DeltaZ % DeltaY) * 2;
        int AdjDown = DeltaY * 2;
        int ErrorTerm = (DeltaZ % DeltaY) - (DeltaY * 2);
        if ((WholeStep & 0x01) != 0)  ErrorTerm += DeltaY;

        do
          {
            int RunLength = WholeStep; 
            if ((ErrorTerm += AdjUp) > 0)
              {
                RunLength++;
                ErrorTerm -= AdjDown; 
              }
            Z0 += ZDir * RunLength;

            Y0++;

            if( (Z = z_buf_->get( X0, Y0 )) < Z0 )
              {
                z_buf_->set( X0, Y0, Z0 );
                point( X0, Y0 );
              }
          }
        while (--DeltaY != 0);
        return;
      }

    /*if( DeltaX == DeltaY ) 
      {
        // Diagonal line
        do 
          {
            X0 += XDir;
            Y0++;
            point( X0, Y0 );
          } 
        while (--DeltaY != 0);
        return;
      }*/

    ErrorAcc = 0;    // initialize the line error accumulator to 0
    Z_Error  = 0;

    // # of bits by which to shift ErrorAcc to get intensity level
    //IntensityShift = 16 - IntensityBits;
    IntensityShift = 16 - 8;

    // Mask used to flip all bits in an intensity weighting, producing the
    // result (1 - intensity weighting)
    //WeightingComplementMask = NumLevels - 1;
    WeightingComplementMask = 256 - 1;

    // Is this an X-major or Y-major line?
    if (DeltaY > DeltaX) 
      {
        // Y-major line; calculate 16-bit fixed-point fractional part of a
        //pixel that X advances each time Y advances 1 pixel, truncating the
        //result so that we won't overrun the endpoint along the X axis
        ErrorAdj = ((unsigned short) DeltaX << 16) / (unsigned short) DeltaY;

        int WholeStep = DeltaZ / DeltaY;
        int AdjUp = (DeltaZ % DeltaY) * 2;
        int AdjDown = DeltaY * 2;
        int ErrorTerm = (DeltaZ % DeltaY) - (DeltaY * 2);
        if ((WholeStep & 0x01) != 0)  ErrorTerm += DeltaY;

        while( --DeltaY )
          {
            int RunLength = WholeStep; 
            if ((ErrorTerm += AdjUp) > 0)
              {
                RunLength++;
                ErrorTerm -= AdjDown; 
              }
            Z0 += ZDir * RunLength;

            ErrorAccTemp = ErrorAcc;   // remember currrent accumulated error
            ErrorAcc += ErrorAdj;      // calculate error for next pixel
            if( ErrorAcc <= ErrorAccTemp )   X0 += XDir;
            Y0++; 

            if( (Z = z_buf_->get( X0, Y0 )) < Z0 )
              {
                z_buf_->set( X0, Y0, Z0 );

                // The IntensityBits most significant bits of ErrorAcc give us the
                // intensity weighting for this pixel, and the complement of the
                // weighting for the paired pixel
                Weighting = ErrorAcc >> IntensityShift;

                unsigned short
                  Shift = (255-us_abs(Weighting  - (Weighting ^ WeightingComplementMask)))>>2;

                Rgb col = pixel_colour( X0 + XDir, Y0 );
                Rbg = red   ( col );
                Gbg = green ( col );
                Bbg = blue  ( col );

                point( X0 + XDir, Y0,
                    (R*(Weighting+Shift)
                    +Rbg*((Weighting ^ WeightingComplementMask)-Shift))>>8,
                    (G*(Weighting+Shift)
                    +Gbg*((Weighting ^ WeightingComplementMask)-Shift))>>8,
                    (B*(Weighting+Shift)
                    +Bbg*((Weighting ^ WeightingComplementMask)-Shift))>>8 );
                point( X0, Y0,
                    (R*((Weighting ^ WeightingComplementMask)+Shift)
                    +Rbg*(Weighting-Shift))>>8,
                    (G*((Weighting ^ WeightingComplementMask)+Shift)
                    +Gbg*(Weighting-Shift))>>8,
                    (B*((Weighting ^ WeightingComplementMask)+Shift)
                    +Bbg*(Weighting-Shift))>>8 );
              }
          }
        if( (Z = z_buf_->get( X1, Y1 )) < Z1 )
          {
            z_buf_->set( X1, Y1, Z1 );
            point( X1, Y1 );
          }
      }
    else
      {
        // It's an X-major line; calculate 16-bit fixed-point fractional part of a
        // pixel that Y advances each time X advances 1 pixel, truncating the
        // result to avoid overrunning the endpoint along the X axis
        ErrorAdj = ((unsigned short) DeltaY << 16) / (unsigned short) DeltaX;

        int WholeStep = DeltaZ / DeltaX;
        int AdjUp = (DeltaZ % DeltaX) * 2;
        int AdjDown = DeltaX * 2;
        int ErrorTerm = (DeltaZ % DeltaX) - (DeltaX * 2);
        if ((WholeStep & 0x01) != 0) ErrorTerm += DeltaX;

        while( --DeltaX )
          {
            int RunLength = WholeStep; 
            if ((ErrorTerm += AdjUp) > 0)
              {
                RunLength++;
                ErrorTerm -= AdjDown; 
              }
            Z0 += ZDir * RunLength;

            ErrorAccTemp = ErrorAcc;   // remember currrent accumulated error
            ErrorAcc += ErrorAdj;      // calculate error for next pixel
            if( ErrorAcc <= ErrorAccTemp )  Y0++;
            X0 += XDir; 

            if( (Z = z_buf_->get( X0, Y0 )) < Z0 )
              {
                z_buf_->set( X0, Y0, Z0 );

                // The IntensityBits most significant bits of ErrorAcc give us the
                // intensity weighting for this pixel, and the complement of the
                // weighting for the paired pixel
                Weighting = ErrorAcc >> IntensityShift;
                //BaseColor + (Weighting ^ WeightingComplementMask));

                unsigned short
                  Shift  = (255-us_abs(Weighting  - (Weighting ^ WeightingComplementMask)))>>2;

                Rgb col = pixel_colour( X0 + XDir, Y0 );
                Rbg = red   ( col );
                Gbg = green ( col );
                Bbg = blue  ( col );

                point( X0, Y0 + 1,
                    (R*(Weighting+Shift)
                    +Rbg*((Weighting ^ WeightingComplementMask)-Shift))>>8,
                    (G*(Weighting+Shift)
                    +Gbg*((Weighting ^ WeightingComplementMask)-Shift))>>8,
                    (B*(Weighting+Shift)
                    +Bbg*((Weighting ^ WeightingComplementMask)-Shift))>>8 );
                point( X0, Y0,
                    (R*((Weighting ^ WeightingComplementMask)+Shift)
                    +Rbg*(Weighting-Shift))>>8,
                    (G*((Weighting ^ WeightingComplementMask)+Shift)
                    +Gbg*(Weighting-Shift))>>8,
                    (B*((Weighting ^ WeightingComplementMask)+Shift)
                    +Bbg*(Weighting-Shift))>>8 );
              }
          }
        if( (Z = z_buf_->get( X1, Y1 )) < Z1 )
          {
            z_buf_->set( X1, Y1, Z1 );
            point( X1, Y1 );
          }
      }
  }
/*
void Painter_3D::
Wu_line2( int X0, int Y0, int Z0, int X1, int Y1, int Z1 )
  {
//    unsigned int  IntensityShift, ErrorAdj, ErrorAcc;
//    unsigned int  ErrorAccTemp, Weighting, WeightingComplementMask;
//    int            DeltaX, DeltaY, Temp, XDir;
    unsigned short  IntensityShift, ErrorAdj, ErrorAcc, Z, Z_Rate, Z_Error;
    unsigned short  ErrorAccTemp, Weighting, WeightingComplementMask;
    short            DeltaX, DeltaY, DeltaZ, Temp, XDir, ZDir;
    Byte            Rbg, Gbg, Bbg;

    // Make sure the line runs top to bottom
    if( Y0 > Y1 )
      {
        Temp = Y0; Y0 = Y1; Y1 = Temp;
        Temp = X0; X0 = X1; X1 = Temp;
      }
      // Draw the initial pixel, which is always exactly intersected by
      // the line and so needs no weighting
      point( X0, Y0 );

    if( (DeltaX = X1 - X0) >= 0 ) 
      {
        XDir = 1;
      } 
    else 
      {
        XDir = -1;
        DeltaX = -DeltaX; // make DeltaX positive 
      }

    if( (DeltaZ = Z1 - Z0) >= 0 )
      {
        ZDir = 1;
      }
    else
      {
        ZDir = -1;
        DeltaZ = -DeltaZ;
      }

    // Special-case horizontal, vertical, and diagonal lines, which
    // require no weighting because they go right through the center of
    // every pixel
    if( (DeltaY = Y1 - Y0) == 0 ) 
      {
        // Horizontal line
        while (DeltaX-- != 0) 
          {
            X0 += XDir;
            point( X0, Y0 );
            point( X0, Y0+1 );
          }
        return;
      }

    if(DeltaX == 0) 
      {
        // Vertical line
        do 
          {
            Y0++;
            point( X0, Y0 );
            point( X0+1, Y0 );
          }
        while (--DeltaY != 0);
        return;
      }

    if( DeltaX == DeltaY ) 
      {
        // Diagonal line
        do 
          {
            X0 += XDir;
            Y0++;
            point( X0, Y0 );
            point( X0-1, Y0 );
          } 
        while (--DeltaY != 0);
        return;
      }

    // Line is not horizontal, diagonal, or vertical

    ErrorAcc = 0;    // initialize the line error accumulator to 0
    
    // # of bits by which to shift ErrorAcc to get intensity level
    //IntensityShift = 16 - IntensityBits;
    IntensityShift = 16 - 8;

    // Mask used to flip all bits in an intensity weighting, producing the
    // result (1 - intensity weighting)
    //WeightingComplementMask = NumLevels - 1;
    WeightingComplementMask = 256 - 1;
    
    // Is this an X-major or Y-major line?
    if (DeltaY > DeltaX) 
      {
        // Y-major line; calculate 16-bit fixed-point fractional part of a
        //pixel that X advances each time Y advances 1 pixel, truncating the
        //result so that we won't overrun the endpoint along the X axis
        ErrorAdj = ((unsigned short) DeltaX << 16) / (unsigned short) DeltaY;

        if( DeltaZ > DeltaY )
          {
            Z_Rate = DeltaZ / DeltaY;
            DeltaZ = DeltaY * Z_Rate;
          }
        else Z_Rate = 0;

        // Draw all pixels other than the first and last
        while( --DeltaY )
          {
            ErrorAccTemp = ErrorAcc;   // remember currrent accumulated error
            ErrorAcc += ErrorAdj;      // calculate error for next pixel
            if( ErrorAcc <= ErrorAccTemp ) 
              {
                // The error accumulator turned over, so advance the X coord
                X0 += XDir;
              }
            Y0++; // Y-major, so always advance Y

            Z0 += Z_Rate;

            if( (Z = z_buf_->get( X0, Y0 )) < Z0 )
              {
                z_buf_->set( X0, Y0, Z0 );

                // The IntensityBits most significant bits of ErrorAcc give us the
                // intensity weighting for this pixel, and the complement of the
                // weighting for the paired pixel
                Weighting = ErrorAcc >> IntensityShift;

                unsigned short
                  Shift  = (255-abs(Weighting  - (Weighting ^ WeightingComplementMask)))>>3;

                get_pixel_colour( X0 + XDir, Y0, &Rbg, &Gbg, &Bbg );

                point( X0 + XDir, Y0,
                    (R_*(Weighting+Shift)
                    +Rbg*((Weighting ^ WeightingComplementMask)-Shift))>>8,
                    (G_*(Weighting+Shift)
                    +Gbg*((Weighting ^ WeightingComplementMask)-Shift))>>8,
                    (B_*(Weighting+Shift)
                    +Bbg*((Weighting ^ WeightingComplementMask)-Shift))>>8 );
                point( X0, Y0, R_, G_, B_ );
                point( X0 - XDir, Y0,
                    (R_*((Weighting ^ WeightingComplementMask)+Shift)
                    +Rbg*(Weighting-Shift))>>8,
                    (G_*((Weighting ^ WeightingComplementMask)+Shift)
                    +Gbg*(Weighting-Shift))>>8,
                    (B_*((Weighting ^ WeightingComplementMask)+Shift)
                    +Bbg*(Weighting-Shift))>>8 );
              }
            if( (Z_Error+=ZDir) > 0 )
              {
                Z_Error -= DeltaY;
                Z0 += ZDir;
              }
          }
        // Draw the final pixel, which is always exactly intersected by the line
        // and so needs no weighting
        point( X1, Y1 );
      }
    else
      {
        // It's an X-major line; calculate 16-bit fixed-point fractional part of a
        // pixel that Y advances each time X advances 1 pixel, truncating the
        // result to avoid overrunning the endpoint along the X axis
        ErrorAdj = ((unsigned short) DeltaY << 16) / (unsigned short) DeltaX;

        if( DeltaZ > DeltaX )
          {
            Z_Rate = DeltaZ / DeltaX;
            DeltaZ = DeltaX * Z_Rate;
          }
        else Z_Rate = 0;

        // Draw all pixels other than the first and last
        while( --DeltaX )
          {
            ErrorAccTemp = ErrorAcc;   // remember currrent accumulated error
            ErrorAcc += ErrorAdj;      // calculate error for next pixel
            if( ErrorAcc <= ErrorAccTemp )
              {
                // The error accumulator turned over, so advance the Y coord
                Y0++;
              }
            X0 += XDir; // X-major, so always advance X
            Z0 += Z_Rate;

            if( (Z = z_buf_->get( X0, Y0 )) < Z0 )
              {
                z_buf_->set( X0, Y0, Z0 );

                // The IntensityBits most significant bits of ErrorAcc give us the
                // intensity weighting for this pixel, and the complement of the
                // weighting for the paired pixel
                Weighting = ErrorAcc >> IntensityShift;
                //BaseColor + (Weighting ^ WeightingComplementMask));

                unsigned short
                  Shift  = (255-abs(Weighting  - (Weighting ^ WeightingComplementMask)))>>3;

                get_pixel_colour( X0 + XDir, Y0, &Rbg, &Gbg, &Bbg );

                point( X0, Y0 + 1,
                    (R_*(Weighting+Shift)
                    +Rbg*((Weighting ^ WeightingComplementMask)-Shift))>>8,
                    (G_*(Weighting+Shift)
                    +Gbg*((Weighting ^ WeightingComplementMask)-Shift))>>8,
                    (B_*(Weighting+Shift)
                    +Bbg*((Weighting ^ WeightingComplementMask)-Shift))>>8 );
                point( X0, Y0, R_, G_, B_ );
                point( X0, Y0 - 1,
                    (R_*((Weighting ^ WeightingComplementMask)+Shift)
                    +Rbg*(Weighting-Shift))>>8,
                    (G_*((Weighting ^ WeightingComplementMask)+Shift)
                    +Gbg*(Weighting-Shift))>>8,
                    (B_*((Weighting ^ WeightingComplementMask)+Shift)
                    +Bbg*(Weighting-Shift))>>8 );
              }
            if( (Z_Error+=ZDir) > 0 )
              {
                Z_Error -= DeltaX;
                Z0 += ZDir;
              }
          }
        // Draw the final pixel, which is always exactly intersected by the line
        //   and so needs no weighting 
        point( X1, Y1 );
      }
  }

void Painter_3D::
Wu_line3( int X0, int Y0, int Z0, int X1, int Y1, int Z1 )
  {
//    unsigned int  IntensityShift, ErrorAdj, ErrorAcc;
//    unsigned int  ErrorAccTemp, Weighting, WeightingComplementMask;
//    int            DeltaX, DeltaY, Temp, XDir;
    unsigned short  IntensityShift, ErrorAdj, ErrorAcc, Z, Z_Rate, Z_Error;
    unsigned short  ErrorAccTemp, Weighting, WeightingComplementMask;
    short            DeltaX, DeltaY, DeltaZ, Temp, XDir, ZDir;
    Byte            Rbg, Gbg, Bbg;

    // Make sure the line runs top to bottom
    if( Y0 > Y1 )
      {
        Temp = Y0; Y0 = Y1; Y1 = Temp;
        Temp = X0; X0 = X1; X1 = Temp;
      }
      // Draw the initial pixel, which is always exactly intersected by
      // the line and so needs no weighting 
      point( X0, Y0 );

    if( (DeltaX = X1 - X0) >= 0 )
      {
        XDir = 1;
      } 
    else 
      {
        XDir = -1;
        DeltaX = -DeltaX; // make DeltaX positive 
      }

    if( (DeltaZ = Z1 - Z0) >= 0 )
      {
        ZDir = 1;
      }
    else
      {
        ZDir = -1;
        DeltaZ = -DeltaZ;
      }

    // Special-case horizontal, vertical, and diagonal lines, which
    // require no weighting because they go right through the center of
    // every pixel
    if( (DeltaY = Y1 - Y0) == 0 ) 
      {
        // Horizontal line
        while (DeltaX-- != 0)
          {
            X0 += XDir;
            point( X0, Y0-1 );
            point( X0, Y0 );
            point( X0, Y0+1 );
          }
        return;
      }

    if(DeltaX == 0)
      {
        // Vertical line
        do 
          {
            Y0++;
            point( X0-1, Y0 );
            point( X0, Y0 );
            point( X0, Y0-1 );
          }
        while (--DeltaY != 0);
        return;
      }

    if( DeltaX == DeltaY ) 
      {
        // Diagonal line
        do 
          {
            X0 += XDir;
            Y0++;
            point( X0-1, Y0 );
            point( X0, Y0 );
            point( X0+1, Y0 );
          } 
        while (--DeltaY != 0);
        return;
      }

    // Line is not horizontal, diagonal, or vertical

    ErrorAcc = 0;    // initialize the line error accumulator to 0
    
    // # of bits by which to shift ErrorAcc to get intensity level
    //IntensityShift = 16 - IntensityBits;
    IntensityShift = 16 - 8;

    // Mask used to flip all bits in an intensity weighting, producing the
    // result (1 - intensity weighting)
    //WeightingComplementMask = NumLevels - 1;
    WeightingComplementMask = 256 - 1;
    
    // Is this an X-major or Y-major line?
    if (DeltaY > DeltaX) 
      {
        // Y-major line; calculate 16-bit fixed-point fractional part of a
        //pixel that X advances each time Y advances 1 pixel, truncating the
        //result so that we won't overrun the endpoint along the X axis
        ErrorAdj = ((unsigned short) DeltaX << 16) / (unsigned short) DeltaY;

        if( DeltaZ > DeltaY )
          {
            Z_Rate = DeltaZ / DeltaY;
            DeltaZ = DeltaY * Z_Rate;
          }
        else Z_Rate = 0;

        // Draw all pixels other than the first and last
        while( --DeltaY ) 
          {
            ErrorAccTemp = ErrorAcc;   // remember currrent accumulated error
            ErrorAcc += ErrorAdj;      // calculate error for next pixel
            if( ErrorAcc <= ErrorAccTemp ) 
              {
                // The error accumulator turned over, so advance the X coord
                X0 += XDir;
              }
            Y0++; // Y-major, so always advance Y
            Z0 += Z_Rate;

            if( (Z = z_buf_->get( X0, Y0 )) < Z0 )
              {
                z_buf_->set( X0, Y0, Z0 );

                // The IntensityBits most significant bits of ErrorAcc give us the
                // intensity weighting for this pixel, and the complement of the
                // weighting for the paired pixel
                Weighting = ErrorAcc >> IntensityShift;

                unsigned short
                  Shift  = (255-abs(Weighting  - (Weighting ^ WeightingComplementMask)))>>3;

                get_pixel_colour( X0 + XDir, Y0, &Rbg, &Gbg, &Bbg );

                point( X0 + XDir + XDir, Y0,
                    (R_*(Weighting+Shift)
                    +Rbg*((Weighting ^ WeightingComplementMask)-Shift))>>8,
                    (G_*(Weighting+Shift)
                    +Gbg*((Weighting ^ WeightingComplementMask)-Shift))>>8,
                    (B_*(Weighting+Shift)
                    +Bbg*((Weighting ^ WeightingComplementMask)-Shift))>>8 );
                point( X0 + XDir, Y0, R_, G_, B_ );
                point( X0,        Y0, R_, G_, B_ );
                point( X0 - XDir, Y0,
                    (R_*((Weighting ^ WeightingComplementMask)+Shift)
                    +Rbg*(Weighting-Shift))>>8,
                    (G_*((Weighting ^ WeightingComplementMask)+Shift)
                    +Gbg*(Weighting-Shift))>>8,
                    (B_*((Weighting ^ WeightingComplementMask)+Shift)
                    +Bbg*(Weighting-Shift))>>8 );
              }
            if( (Z_Error+=ZDir) > 0 )
              {
                Z_Error -= DeltaY;
                Z0 += ZDir;
              }
          }
        // Draw the final pixel, which is always exactly intersected by the line
        // and so needs no weighting
        point( X1, Y1 );
      }
    else
      {
        // It's an X-major line; calculate 16-bit fixed-point fractional part of a
        // pixel that Y advances each time X advances 1 pixel, truncating the
        // result to avoid overrunning the endpoint along the X axis
        ErrorAdj = ((unsigned short) DeltaY << 16) / (unsigned short) DeltaX;

        if( DeltaZ > DeltaX )
          {
            Z_Rate = DeltaZ / DeltaX;
            DeltaZ = DeltaX * Z_Rate;
          }
        else Z_Rate = 0;

        // Draw all pixels other than the first and last
        while( --DeltaX )
          {
            ErrorAccTemp = ErrorAcc;   // remember currrent accumulated error
            ErrorAcc += ErrorAdj;      // calculate error for next pixel
            if( ErrorAcc <= ErrorAccTemp )
              {
                // The error accumulator turned over, so advance the Y coord
                Y0++;
              }
            X0 += XDir; // X-major, so always advance X
            Z0 += Z_Rate;

            if( (Z = z_buf_->get( X0, Y0 )) < Z0 )
              {
                z_buf_->set( X0, Y0, Z0 );

                // The IntensityBits most significant bits of ErrorAcc give us the
                // intensity weighting for this pixel, and the complement of the
                // weighting for the paired pixel
                Weighting = ErrorAcc >> IntensityShift;
                //BaseColor + (Weighting ^ WeightingComplementMask));

                unsigned short
                  Shift  = (255-abs(Weighting  - (Weighting ^ WeightingComplementMask)))>>3;

                get_pixel_colour( X0 + XDir, Y0, &Rbg, &Gbg, &Bbg );

                point( X0, Y0 + 2,
                    (R_*(Weighting+Shift)
                    +Rbg*((Weighting ^ WeightingComplementMask)-Shift))>>8,
                    (G_*(Weighting+Shift)
                    +Gbg*((Weighting ^ WeightingComplementMask)-Shift))>>8,
                    (B_*(Weighting+Shift)
                    +Bbg*((Weighting ^ WeightingComplementMask)-Shift))>>8 );
                point( X0, Y0 + 1, R_, G_, B_ );
                point( X0, Y0,      R_, G_, B_ );
                point( X0, Y0 - 1,
                    (R_*((Weighting ^ WeightingComplementMask)+Shift)
                    +Rbg*(Weighting-Shift))>>8,
                    (G_*((Weighting ^ WeightingComplementMask)+Shift)
                    +Gbg*(Weighting-Shift))>>8,
                    (B_*((Weighting ^ WeightingComplementMask)+Shift)
                    +Bbg*(Weighting-Shift))>>8 );
          }
            if( (Z_Error+=ZDir) > 0 )
              {
                Z_Error -= DeltaX;
                Z0 += ZDir;
              }
          }
        // Draw the final pixel, which is always exactly intersected by the line
        //   and so needs no weighting
        point( X1, Y1 );
      }
  }
*/
//------- sqrt_tab ------------------------

const int MAX_RAD = 150;                                        //fix all

extern int sqrt_tab[ 100000 ];
int sqrt_tab[ 100000 ];

struct Sqrt_tab_init
{
    Sqrt_tab_init()
    {
        for( int i = 0;  i < 100000; i++ )
        sqrt_tab[ i ] = ( int )( sqrt( ( double )( i << 4 ) ) );
    }
}

init_sqrt_tab;


//-----------------------------------------

#define GETCOLOR                                                            \
   {                                                                        \
   color_flag = 0;                                                          \
   cos_theta = ((int)(dZ*R_1*255));                                         \
   D = Z_clip - Z;                                                          \
   if( D < 0 )                                                              \
   {                                                                        \
   D = R;                                                                   \
   cos_theta = 220;                                                         \
   if( Z_clip - (Z0 - dZ) < 0 ) continue;                                   \
                                                                            \
   light_model_.I( &red,&green,&blue, &Material, cos_theta, D >> shift );   \
   red = (red+765)>>2; green = (green+765)>>2; blue = (blue+600)>>2;        \
   }                                                                        \
   else                                                                     \
   light_model_.I( &red,&green,&blue, &Material, cos_theta, D >> shift );   \
   }

void Painter_3D::
sphere( View_abstract_sphere * sphere, int Z_clip )
{
    int     incr = ( shift == 0 ) ? 1 : 2 << ( shift - 1 );
    int     D, Z, x, y, x0, y0, dx, dy, dX, dY, dZ,  dY2;
    int     cos_theta;
    int     color_flag;

    int     red, green, blue;
    //Byte      red,green,blue;
    Material  Material = sphere->material_;

    int     X0 = sphere->X_ << shift,
            Y0 = sphere->Y_ << shift,
            Z0 = sphere->Z_ << shift;
    int     R  = int( sphere->radius_ ) << shift;
    double  R_1 = 1.0 / R;
    int     r2, R2 = R * R;

    Z_clip += shift; //fix ???

    Z_clip <<= shift;

    if(     sphere->X_ - sphere->radius_ > 0
        &&  sphere->X_ + sphere->radius_ < width()
        &&  sphere->Y_ - sphere->radius_ > 0
        &&  sphere->Y_ + sphere->radius_ < height()  )
    {
        for( dY = 0; dY < R; dY += incr )
        {
            dY2 = dY * dY;

            for( dX = dY; ; dX += incr )
            {
                r2 = dX * dX + dY2;
                if( r2 >= R2 )  break;

                //dZ = int(sqrt( R2 - r2 ));
                dZ = sqrt_tab[ ( R2 - r2 ) >> 4 ];

                Z = Z0 + dZ;
                cos_theta = ( ( int )( dZ * R_1 * 255 ) );
                D = Z_clip - Z;

                /*if( D < 0 )   //  hollow
                {
                    Z = Z0 - dZ;
                    D = Z_clip - Z;
                    if( Z_clip - Z < 0 ) continue;
                }*/

                //if( D < 0 )   continue;  //  half
                if( D < 0 ) // solid
                {
                    //Z = Z_clip;
                    D = R;
                    cos_theta = 220;
                    if( Z_clip - ( Z0 - dZ ) < 0 ) continue;

                    light_model_.I( & red, & green, & blue, & Material,
                                    cos_theta, D >> shift );
                    //                    red = (red+1600)>>3; green =
                    // (green+1600)>>3; blue = (blue+1400)>>3;
                    red = ( red + 765 ) >> 2;
                    green = ( green + 765 ) >> 2;
                    blue = ( blue + 600 ) >> 2;
                    point( red, green, blue );
                }

                /*if( D < 0 )
                {
                    red = 180; green = 180; blue = 160;
                }        */

                else
                {
                    light_model_.I( & red, & green, & blue, & Material,
                                    cos_theta, D >> shift );
                    point( red, green, blue );
                }

                dx = dX >> shift;
                dy = dY >> shift;
                x0 = X0 >> shift;
                y0 = Y0 >> shift;

                x  = x0 + dx;
                y  = y0 + dy;

/*                __asm
                {
                  nop
                  mov eax,eax
                  nop
                }    //*/

                if( Z >= z_buf_->get( x, y ) )        // 1 sector
                {
                    z_buf_->set( x, y, Z );
                    //point( x, y, red,green,blue );
                    point( x, y );
                } //*/

/*__asm
{
//125:                                  if( Z >= z_buf->get( x, y ) )               // 1 sector
mov         esi,dword ptr [this]
mov         eax,dword ptr [esi+30h]
mov         ecx,dword ptr [x]
mov         eax,dword ptr [eax]
mov         eax,dword ptr [eax+ecx*4]

mov         ecx,dword ptr [y]
lea         eax,dword ptr [eax+ecx*4]
mov         ecx,dword ptr [Z]

cmp         ecx,dword ptr [eax]

jl          END
//126:                                      {
//127:                                          z_buf->set( x, y, Z );
mov         dword ptr [eax],ecx

//128:                                          point( x, y, red,green,blue );
mov         eax,dword ptr [esi+0Ch]
mov         ecx,dword ptr [y]
sub         eax,ecx

mov         ecx,dword ptr [x]
imul        eax,dword ptr [esi+8]
lea         ecx,dword ptr [ecx+ecx*2]
add         eax,dword ptr [esi+4]
//mov         edi,dword ptr [esi+4]
//add         eax,edi
add         eax,ecx

mov         dl,byte ptr [blue]
mov         byte ptr [eax],dl
mov         dl,byte ptr [green]
mov         byte ptr [eax+1],dl
mov         dl,byte ptr [red]
mov         byte ptr [eax+2],dl

END:
}      //*/

                /*__asm
                {
                  nop
                  mov eax,eax
                  nop
                }   //*/

                y = ( y0 - dy );
                if( Z > z_buf_->get( x, y ) )       // 8 sector
                {
                    z_buf_->set( x, y, Z );
                    //point( x, y, red,green,blue );
                    point( x, y );
                }

                x = ( x0 - dx );
                if( Z > z_buf_->get( x, y ) )       // 5 sector
                {
                    z_buf_->set( x, y, Z );
                    //point( x, y, red,green,blue );
                    point( x, y );
                }

                y = ( y0 + dy );
                if( Z > z_buf_->get( x, y ) )       // 4 sector
                {
                    z_buf_->set( x, y, Z );
                    //point( x, y, red,green,blue );
                    point( x, y );
                }

                x = ( x0 + dy );
                y = ( y0 + dx );
                if( Z > z_buf_->get( x, y ) )       // 2 sector
                {
                    z_buf_->set( x, y, Z );
                    //point( x, y, red,green,blue );
                    point( x, y );
                }

                x = ( x0 - dy );
                if( Z > z_buf_->get( x, y ) )       // 3 sector
                {
                    z_buf_->set( x, y, Z );
                    //point( x, y, red,green,blue );
                    point( x, y );
                }

                y = ( y0 - dx );
                if( Z > z_buf_->get( x, y ) )       // 6 sector
                {
                    z_buf_->set( x, y, Z );
                    //point( x, y, red,green,blue );
                    point( x, y );
                }

                x = ( x0 + dy );
                if( Z > z_buf_->get( x, y ) )       // 7 sector
                {
                    z_buf_->set( x, y, Z );
                    //point( x, y, red,green,blue );
                    point( x, y );
                } // */

                /*asm__
                {
                }  */
            }
        }
    }
    else
    {
        int width  = this->width(), //fix
            height = this->height();

        for( dY = 0; dY < R; dY += incr )
        {
            dY2 = dY * dY;

            //for( dx = dy; ;dx+=2<<(shift-1) )
            for( dX = dY; ; dX += incr )
            {
                r2 = dX * dX + dY2;
                if( r2 >= R2 )  break;

                dZ = int( sqrt( ( double )( R2 - r2 ) ) );
                //dZ = sqrt_tab[ R2 - r2 ];

                Z = Z0 + dZ;

                //dx = dX >> shift;    dy = dY >> shift;

                x = ( X0 + dX ) >> shift;
                y = ( Y0 + dY ) >> shift;
                //x = (x0+dx);    y = (y0+dy);
                color_flag = 1;

                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );

//                        if( color_flag )  GETCOLOR
//                            point( x, y, red, green, blue );
                        if( color_flag )  
    {                                                                        
        color_flag = 0;                                                          
        cos_theta = ((int)(dZ*R_1*255));                                         
        D = Z_clip - Z;                                                          
        if( D < 0 )                                                              
        {                                                                        
            D = R;                                                                   
            cos_theta = 220;                                                         
            if( Z_clip - (Z0 - dZ) < 0 ) continue;                                   
                                                                            
            light_model_.I( &red,&green,&blue, &Material, cos_theta, D >> shift );   
            red = (red+765)>>2; green = (green+765)>>2; blue = (blue+600)>>2;        
        }                                                                        
        else                                                                     
        light_model_.I( &red,&green,&blue, &Material, cos_theta, D >> shift );   
    }
                            
                            point( x, y, red, green, blue );
                    }

                y = ( Y0 - dY ) >> shift;
                //y = (y0-dy);
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );

                        if( color_flag )  GETCOLOR
                            point( x, y, red, green, blue );
                    }

                x = ( X0 - dX ) >> shift;
                //x = (x0-dx);
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );

                        if( color_flag )  GETCOLOR
                            point( x, y, red, green, blue );
                    }

                y = ( Y0 + dY ) >> shift;
                //y = (y0+dy);
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );

                        if( color_flag )  GETCOLOR
                            point( x, y, red, green, blue );
                    }

                x = ( X0 + dY ) >> shift;
                y = ( Y0 + dX ) >> shift;
                //x = (x0+dy);    y = (y0+dx);
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );

                        if( color_flag )  GETCOLOR
                            point( x, y, red, green, blue );
                    }

                x = ( X0 - dY ) >> shift;
                //x = (x0-dy);
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );

                        if( color_flag )  GETCOLOR
                            point( x, y, red, green, blue );
                    }

                y = ( Y0 - dX ) >> shift;
                //y = (y0-dx);
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );

                        if( color_flag )  GETCOLOR
                            point( x, y, red, green, blue );
                    }

                x = ( X0 + dY ) >> shift;
                //x = (x0+dy);
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );

                        if( color_flag )  GETCOLOR
                            point( x, y, red, green, blue );
                    }
            }
        }
    }
}

void Painter_3D::
crude_sphere( View_abstract_sphere * sphere, int Z_clip, int resolution )
{
    int      incr = ( ( shift == 0 ) ? 1 : 2 << ( shift - 1 ) ) << resolution;
    int      D, Z, x, y, x0, y0, dx, dy, dX, dY, dZ,  dY2;
    int      cos_theta;
    int     color_flag;
    int      red, green, blue;
    Material  Material = sphere->material_;

    int      X0 = sphere->X_ << shift, Y0 = sphere->Y_ << shift, Z0 =
       sphere->Z_ << shift;
    int      R  = int( sphere->radius_ ) << shift;
    double  R_1 = 1.0 / R;
    int      r2, R2 = R * R;
    Z_clip <<= shift;

    if( sphere->X_ - sphere->radius_ - ( 1 << resolution ) > 0 &&
    sphere->X_ + sphere->radius_ + ( 1 << resolution ) < width() &&
    sphere->Y_ - sphere->radius_ - ( 1 << resolution ) > 0 &&
    sphere->Y_ + sphere->radius_ + ( 1 << resolution ) < height()  )
    {
        for( dY = 0; dY < R; dY += incr )
        {
            dY2 = dY * dY;

            for( dX = dY; ; dX += incr )
            {
                r2 = dX * dX + dY2;
                if( r2 >= R2 )  break;

                //dZ = int(sqrt( R2 - r2 ));
                dZ = sqrt_tab[ ( R2 - r2 ) >> 4 ];

                Z = Z0 + dZ;
                cos_theta = ( ( int )( dZ * R_1 * 255 ) );
                D = Z_clip - Z;

                GETCOLOR

                dx = dX >> shift;
                dy = dY >> shift;

                x0 = X0 >> shift;
                y0 = Y0 >> shift;

                x0 = ( x0 >> resolution ) << resolution;
                dx = ( dx >> resolution ) << resolution;
                y0 = ( y0 >> resolution ) << resolution;
                dy = ( dy >> resolution ) << resolution;

                x  = x0 + dx;
                y  = y0 + dy;

                if( Z >= z_buf_->get( x, y ) )        // 1 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, resolution );
                }

                y = ( y0 - dy );
                if( Z > z_buf_->get( x, y ) )       // 8 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, resolution );
                }

                x = ( x0 - dx );
                if( Z > z_buf_->get( x, y ) )       // 5 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, resolution );
                }

                y = ( y0 + dy );
                if( Z > z_buf_->get( x, y ) )       // 4 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, resolution );
                }

                x = ( x0 + dy );
                y = ( y0 + dx );
                if( Z > z_buf_->get( x, y ) )       // 2 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, resolution );
                }

                x = ( x0 - dy );
                if( Z > z_buf_->get( x, y ) )       // 3 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, resolution );
                }

                y = ( y0 - dx );
                if( Z > z_buf_->get( x, y ) )       // 6 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, resolution );
                }

                x = ( x0 + dy );
                if( Z > z_buf_->get( x, y ) )       // 7 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, resolution );
                } // */

                /*asm__
                {
                }  */
            }
        }
    }
    else
    {
        int width = int( this->width() ) - ( incr >> shift ),
           height = int( this->height() ) - ( incr >> shift );

        for( dY = 0; dY < R; dY += incr )
        {
            dY2 = dY * dY;

            //for( dx = dy; ;dx+=2<<(shift-1) )
            for( dX = dY; ; dX += incr )
            {
                r2 = dX * dX + dY2;
                if( r2 >= R2 )  break;

                dZ = int( sqrt( ( double )( R2 - r2 ) ) );
                //dZ = sqrt_tab[ R2 - r2 ];

                Z = Z0 + dZ;

                GETCOLOR

                dx = dX >> shift;
                dy = dY >> shift;
                x0 = X0 >> shift;
                y0 = Y0 >> shift;

                x0 = ( x0 >> resolution ) << resolution;
                dx = ( dx >> resolution ) << resolution;
                y0 = ( y0 >> resolution ) << resolution;
                dy = ( dy >> resolution ) << resolution;

                x = ( x0 + dx );
                y = ( y0 + dy );

                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue, resolution );
                    }

                y = ( y0 - dy );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue, resolution );
                    }

                x = ( x0 - dx );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue, resolution );
                    }

                y = ( y0 + dy );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue, resolution );
                    }

                x = ( x0 + dy );
                y = ( y0 + dx );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue, resolution );
                    }

                x = ( x0 - dy );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue, resolution );
                    }

                y = ( y0 - dx );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue, resolution );
                    }

                x = ( x0 + dy );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue, resolution );
                    }
            }
        }
    }
}

void Painter_3D::
crude_sphere2( View_abstract_sphere * sphere, int Z_clip, int resolution )
{
    int      incr = ( ( shift == 0 ) ? 1 : 2 << ( shift - 1 ) ) << resolution;
    int      D, Z, x, y, x0, y0, dx, dy, dX, dY, dZ,  dY2;
    int      cos_theta;
    int     color_flag;
    int      corrected_resolution = resolution - 1;

    int      red, green, blue;
    Material  Material = sphere->material_;

    int      X0 = sphere->X_ << shift, Y0 = sphere->Y_ << shift, Z0 =
       sphere->Z_ << shift;
    int      R  = int( sphere->radius_ ) << shift;
    double  R_1 = 1.0 / R;
    int      r2, R2 = R * R;
    Z_clip <<= shift;

    if( sphere->X_ - sphere->radius_ - ( 1 << resolution ) > 0 &&
    sphere->X_ + sphere->radius_ + ( 1 << resolution ) < width() &&
    sphere->Y_ - sphere->radius_ - ( 1 << resolution ) > 0 &&
    sphere->Y_ + sphere->radius_ + ( 1 << resolution ) < height()  )
    {
        for( dY = 0; dY < R; dY += incr )
        {
            dY2 = dY * dY;

            for( dX = dY; ; dX += incr )
            {
                r2 = dX * dX + dY2;
                if( r2 >= R2 )  break;

                //dZ = int(sqrt( R2 - r2 ));
                dZ = sqrt_tab[ ( R2 - r2 ) >> 4 ];

                Z = Z0 + dZ;
                cos_theta = ( ( int )( dZ * R_1 * 255 ) );
                D = Z_clip - Z;

                GETCOLOR

                dx = dX >> shift;
                dy = dY >> shift;

                x0 = X0 >> shift;
                y0 = Y0 >> shift;

                x0 = ( x0 >> resolution ) << resolution;
                dx = ( dx >> resolution ) << resolution;
                y0 = ( y0 >> resolution ) << resolution;
                dy = ( dy >> resolution ) << resolution;

                x  = x0 + dx;
                y  = y0 + dy;

                if( Z >= z_buf_->get( x, y ) )        // 1 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, corrected_resolution );
                }

                y = ( y0 - dy );
                if( Z > z_buf_->get( x, y ) )       // 8 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, corrected_resolution );
                }

                x = ( x0 - dx );
                if( Z > z_buf_->get( x, y ) )       // 5 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, corrected_resolution );
                }

                y = ( y0 + dy );
                if( Z > z_buf_->get( x, y ) )       // 4 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, corrected_resolution );
                }

                x = ( x0 + dy );
                y = ( y0 + dx );
                if( Z > z_buf_->get( x, y ) )       // 2 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, corrected_resolution );
                }

                x = ( x0 - dy );
                if( Z > z_buf_->get( x, y ) )       // 3 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, corrected_resolution );
                }

                y = ( y0 - dx );
                if( Z > z_buf_->get( x, y ) )       // 6 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, corrected_resolution );
                }

                x = ( x0 + dy );
                if( Z > z_buf_->get( x, y ) )       // 7 sector
                {
                    z_buf_->set( x, y, Z );
                    square( x, y, red, green, blue, corrected_resolution );
                } // */

                /*asm__
                {
                }  */
            }
        }
    }
    else
    {
        int width = int( this->width() ) - ( incr >> shift ),
           height = int( this->height() ) - ( incr >> shift );

        for( dY = 0; dY < R; dY += incr )
        {
            dY2 = dY * dY;

            //for( dx = dy; ;dx+=2<<(shift-1) )
            for( dX = dY; ; dX += incr )
            {
                r2 = dX * dX + dY2;
                if( r2 >= R2 )  break;

                dZ = int( sqrt( ( double )( R2 - r2 ) ) );
                //dZ = sqrt_tab[ R2 - r2 ];

                Z = Z0 + dZ;

                GETCOLOR

                dx = dX >> shift;
                dy = dY >> shift;
                x0 = X0 >> shift;
                y0 = Y0 >> shift;

                x0 = ( x0 >> resolution ) << resolution;
                dx = ( dx >> resolution ) << resolution;
                y0 = ( y0 >> resolution ) << resolution;
                dy = ( dy >> resolution ) << resolution;

                x = ( x0 + dx );
                y = ( y0 + dy );

                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue,
                           corrected_resolution );
                    }

                y = ( y0 - dy );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue,
                           corrected_resolution );
                    }

                x = ( x0 - dx );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue,
                           corrected_resolution );
                    }

                y = ( y0 + dy );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue,
                           corrected_resolution );
                    }

                x = ( x0 + dy );
                y = ( y0 + dx );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue,
                           corrected_resolution );
                    }

                x = ( x0 - dy );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue,
                           corrected_resolution );
                    }

                y = ( y0 - dx );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue,
                           corrected_resolution );
                    }

                x = ( x0 + dy );
                if( x > 0  &&  x < width  &&  y > 0  &&  y < height )
                    if( Z > z_buf_->get( x, y ) )
                    {
                        z_buf_->set( x, y, Z );
                        square( x, y, red, green, blue,
                           corrected_resolution );
                    }
            }
        }
    }
}

void Painter_3D::clear()
{
    Painter_2D::clear();
    z_buf_->clear();
}

