/* 
    This file is part of the ChemCon molecular mechanics 
    and molecular graphics program.

    Copyright (C) 1997-2002  Alexei Nikitin 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "__rand.h"

const unsigned  long  __RAND_MAX__ =  4294967295UL;

//----------------------  Base_Randoms -------------------------------------------------------

#ifndef INCLUDED_time
  #include <time.h>
  #define INCLUDED_time
#endif

unsigned  long  T_Base_random::
calc()
  {
    int    j;
    switch( method )
      {
        case WATCOM  :
                        value =  value * 1103515245UL + 12345UL ;
                        break;
        case Knuth  :
                        value =  value * 3141592621UL + 907633385UL;
                        break;
        case Bad    :
                        value =  value * 11UL;
                        break;
        case Mix    :
                        valueX = valueX * 1103515245UL + 12345UL ;
                        valueY = valueY * 3141592621UL + 907633385UL;
                        j = int( T_Base_random__TABLE_SIZE * valueY /  __RAND_MAX__ );
                        value   =  table[ j ];
                        table[ j ] = valueX;
                        break;
        case Koveu  :
                        value =  value * ( value + 1 );
                        break;
      }
    return      value ; //>> 16;
  }

double  T_Base_random::
next( void )
  {
    return  rmin + ( rmax - rmin ) * double(calc()) / double( __RAND_MAX__ );
  }

double  T_Base_random::
next( double __rmax )
  {
    return  __rmax  * double(calc()) / double( __RAND_MAX__ );
  }

double  T_Base_random::
next( double __rmin, double __rmax )
  {
    return  __rmin + ( __rmax - __rmin ) * double(calc()) / double( __RAND_MAX__ );
  }

void  T_Base_random::
srand( unsigned long seed)
  {
    value = seed;
    switch( method )
      {
        case  Mix    :    valueX = value;    fill_table();    break;
        case  Koveu  :    value = value * 4 + 2;
	default: ; //fix
      }
  }

void  T_Base_random::
randomize( void )
  {
    time_t t;
    time( &t );
    value = t;
    switch( method )
      {
        case  Mix    :    valueX = value;    fill_table();    break;
        case  Koveu  :    value = value * 4 + 2;
	default: ; //fix
      }
  }

void  T_Base_random::
set_method( Methods m )
  {
    switch( m )
      {
        case    Mix    :    fill_table();    break;
        case    Koveu  :    value = 6;      break;
        case    Bad    :    value = 12;      break;
	default: ; //fix
      }
    method = m;
  }

void  T_Base_random::
set_range( double __rmin, double __rmax )
  {
     rmin = __rmin;  rmax = __rmax;
  }

 void  T_Base_random::
 fill_table( void )
  {
    method = WATCOM;
    for( int i=0;  i<T_Base_random__TABLE_SIZE;  i++ )    table[ i ] = calc();
    method = Mix;
  }

T_Base_random::T_Base_random()
  : value( 1 ), valueX( 1 ), valueY( 1 ), rmin( 0 ),  rmax( __RAND_MAX__ ), method( WATCOM ) {  }

T_Base_random::T_Base_random( double __rmax )
  : value( 1 ), valueX( 1 ), valueY( 1 ), rmin( 0 ),  rmax( __rmax ), method( WATCOM )        {  }

T_Base_random::T_Base_random( double __rmin, double __rmax )
  : value( 1 ), valueX( 1 ), valueY( 1 ), rmin( __rmin ), rmax( __rmax ), method( WATCOM )    {  }
        
//----------------------- end  __rand.cpp ---------------------------------------------------------------
