#include "OpenRM_atom.h"

#include "View_atom_type.h"
#include "QRMWidget.h"
#include "Flaw.h"

#include <rm/rm.h>

namespace MM
{

OpenRM_atom::
OpenRM_atom (RMnode *parent_node, QRMWidget *rm_widget)
:
    sphere_(0),
    atom_node_(0),
    parent_node_(parent_node),
    rm_widget_(rm_widget),
    show_flag_  (true)
{
    REQUIRE ("", parent_node != 0);
    REQUIRE ("", rm_widget   != 0);
}

Visual_atom * OpenRM_atom::
clone () const
{
    OpenRM_atom *result = new OpenRM_atom (parent_node_, rm_widget_);

    RMvertex3D  vertex;    vertex.x = 0;    vertex.y = 0;    vertex.z = 0;
    float       radius = .5F;
    RMcolor3D   color  = {1.f, .0f, 1.f};
    
    result->sphere_ = rmPrimitiveNew (RM_SPHERES);

    rmPrimitiveSetVertex3D (result->sphere_, 1, &vertex, RM_COPY_DATA, NULL);
    rmPrimitiveSetRadii    (result->sphere_, 1, &radius, RM_COPY_DATA, NULL);
    rmPrimitiveSetColor3D  (result->sphere_, 1, &color,  RM_COPY_DATA, NULL);

    result->atom_node_ = rmNodeNew ("Atom",  RM_RENDERPASS_ALL, RM_RENDERPASS_ALL);

    //rmPrimitiveSetModelFlag (result->sphere_, RM_SPHERES_32);    
    rmPrimitiveSetModelFlag (result->sphere_, RM_SPHERES_128);   
    //rmPrimitiveSetModelFlag (result->sphere_, RM_SPHERES_512);

    rmNodeAddPrimitive (result->atom_node_, result->sphere_);          
    result->show_flag_ = show_flag_;
    if (show_flag_ == true)
        rmNodeAddChild (result->parent_node_, result->atom_node_);

    ENSURE ("", result->sphere_      != 0);
    ENSURE ("", result->atom_node_   != 0);
    ENSURE ("", result->parent_node_ != 0);
    ENSURE ("", result->rm_widget_   != 0);
    return result;

    //rmNodeSetPreTraversalCallback (atom_node_,RM_VIEW, rmNodeFrustumCullCallback);//*/
}

void OpenRM_atom::
show () const
{
    if (show_flag_ == false) //fix to contract
    {
        rmNodeAddChild (parent_node_, atom_node_);
        show_flag_ = true;
        update ();
    }
}

void OpenRM_atom::
hide () const
{
    if (show_flag_ == true) //fix to contract
    {
        rmNodeRemoveChild (parent_node_, atom_node_);
        show_flag_ = false;
        update ();
    }
}

OpenRM_atom::
~OpenRM_atom ()
{
    RMenum result;

    // for prototype:
    if (atom_node_ == 0)
        return;

    show ();

//    result = rmNodeRemoveAllPrims (atom_node_); //fix remove
//    if (result == RM_WHACKED)
//        FLAW ("rmNodeRemoveAllPrims");

    result = rmNodeRemoveChild (parent_node_, atom_node_);
    if (result == RM_WHACKED)
        FLAW ("rmNodeRemoveChild");

    result = rmSubTreeDelete (atom_node_);
    if (result == RM_WHACKED)
        FLAW ("rmSubTreeDelete");
    
    show_flag_ = false;
    update ();
}

void OpenRM_atom::
update () const
{
    rm_widget_->update ();
}

void OpenRM_atom::
update_coordinates (double x, double y, double z)
{
    RMvertex3D  vertex;    
    vertex.x = float(x);    vertex.y = float(y);    vertex.z = float(z);
    rmPrimitiveSetVertex3D (sphere_, 1, &vertex, RM_COPY_DATA, NULL);
    rmNodeComputeBoundingBox (atom_node_);//fix
    update ();
}

void OpenRM_atom::
update_view_type (View_atom_type & type)
{
    RMcolor3D color;
    color.r       = float (type.r() / 255.0);
    color.g       = float (type.g() / 255.0);
    color.b       = float (type.b() / 255.0);
    float radius  = float (type.radius());
    
    rmPrimitiveSetColor3D (sphere_, 1, &color,  RM_COPY_DATA, NULL);
    rmPrimitiveSetRadii   (sphere_, 1, &radius, RM_COPY_DATA, NULL);
    update ();
}

/*void OpenRM_atom::
get_color (int *R, int *G, int *B)
{
    *R = color.r * 255.;
    *G = color.g * 255.;
    *B = color.b * 255.;
}//*/

void OpenRM_atom::
set_color (int R, int G, int B)
{
    RMcolor3D color;
    color.r       = float (R / 255.0);
    color.g       = float (G / 255.0);
    color.b       = float (B / 255.0);
    
    rmPrimitiveSetColor3D (sphere_, 1, &color,  RM_COPY_DATA, NULL);
    update ();
}

void OpenRM_atom::
set_level_of_details (int lod)
{
    switch (lod)
    {
    case 0: rmPrimitiveSetModelFlag (sphere_, RM_SPHERES_8);
            break;
    case 1: rmPrimitiveSetModelFlag (sphere_, RM_SPHERES_32);
            break;
    case 2: rmPrimitiveSetModelFlag (sphere_, RM_SPHERES_128);
            break;
    case 3: rmPrimitiveSetModelFlag (sphere_, RM_SPHERES_512);
            break;
    default: rmPrimitiveSetModelFlag (sphere_, RM_SPHERES_512);
    }
}

}//MM


