// 
// This file is part of MolMeccano, a cross-platform C++ chemical library.
//
// Copyright (C) 2001,2002 Alexei Nikitin
//
#if defined WIN32
#pragma warning(disable : 4786)
#endif

#include "Test.h"

#include "Component.h"
#include "Model.h"
#include "Text_view.h"
#include "HyperChem_like_builder.h"
#include "ID.h"
#include "Atom.h"
#include "Qt_text_output.h"
#include "Path.h"
#include "Bond.h"
#include "Create.h"


namespace MM
{

class Build_test : public Test
{
public:
    Build_test( const Text & suite_name ) : Test( suite_name ) { }

    void run()
    {
        test1_build_methane();                          count_tests();
        test2_build_methane_from_hyperchem_script();    count_tests();
    }

    void test1_build_methane();
    void test2_build_methane_from_hyperchem_script();

private:
    bool check_methane_model (const Model & model) const; 
};

Build_test test_build("temporary");


Qt_text_output & qt_text_output();

Qt_text_output & qt_text_output()           //fix to console
{
    static Qt_text_output *singleton = new Qt_text_output(); 
    return *singleton;
}


void Build_test::test1_build_methane()
{
    // Create component
    //Own< Component > component( prototype<Component>().clone() );

    //MM_model & model = *prototype <MM_model> ().clone();
    //component().set_model( &model );
    Own <Model> model (prototype <Model>().clone());


//    View & view  = *new Text_view (model(), model(), qt_text_output());
    //component().set_view( &view );
    qt_text_output().show();

    HyperChem_like_builder & model_builder = 
        * new HyperChem_like_builder (model());
    //component().set_controller( &model_builder );


    // Create methane
    // Model::Accessor_protocol & get_model = model;

    const ID & C = model_builder.create_atom( Element(6) );
    
    const ID & H1 = model_builder.create_atom( Element(1) );
    model_builder.set_bond( C, H1, Order('s') );
    
    const ID & H2 = model_builder.create_atom( Element(1) );
    model_builder.set_bond( C, H2, Order('s') );
    
    const ID & H3 = model_builder.create_atom( Element(1) );
    model_builder.set_bond( C, H3, Order('s') );
    
    const ID & H4 = model_builder.create_atom( Element(1) );
    model_builder.set_bond( C, H4, Order('s') );

    const ID & H5 = model_builder.create_atom( Element(1) );
    model().atom( H5 ).erase(); //fix Text_view presentation


    model().atom( C ).set_x( 0 );
    model().atom( C ).set_y( 0 );
    model().atom( C ).set_z( 0 );

    model().atom( H1 ).set_x(-1 );
    model().atom( H1 ).set_y( 0 );
    model().atom( H1 ).set_z( 1 );

    model().atom( H2 ).set_x(-1 );
    model().atom( H2 ).set_y( 0 );
    model().atom( H2 ).set_z(-1 );

    model().atom( H3 ).set_x( 1 );
    model().atom( H3 ).set_y(-1 );
    model().atom( H3 ).set_z( 0 );

    model().atom( H4 ).set_x( 1 );
    model().atom( H4 ).set_y( 1 );
    model().atom( H4 ).set_z( 0 );

    
    //model_builder.setup_3d_geometry();
    //model.save();
    model().save_as (Path::tests_rubbish() + "methane.xyz");
    //fix model.save_as( Text(DIR_tests_rubbish) + "methane.hin" );

    TEST( "", check_methane_model (model()));//*/
}

void Build_test::test2_build_methane_from_hyperchem_script()
{
    /*// Create component
    std::auto_ptr< Component > component = factory().make_component();

    MM_model & model = *factory().make_molecular_model();
    component->set_model( &model );

    //View       & view       = *factory().make_hin_file()
    View & view       = *factory().make_stream_view( std::cout );
    component->set_view( &view );      

    HyperChem_like_script_builder & model_builder = 
        *factory().make_hyperchem_like_script_builder( model );
    component->set_controller( &model_builder );

    // Create methane
    model_builder.execute( "methane.scr" );

    
    TEST( "", check_methane_model( model ) );//*/
}

bool Build_test::
check_methane_model (const Model & model) const 
{
    bool result = true;

    result &= ( model.atom_count() == 5 );  //fix text view - 5 H
    result &= ( model.bond_count() == 4 );
    //result &= ( model.groups().size   () == 0 );
    //result &= ( model.molecules().size() == 1 );
    //result &= ( model.residues().size () == 0 );

    return result;
}//*/

}//MM
