#ifndef GEOMETRY_H
#define GEOMETRY_H

#ifndef USER_H
#include "User.h"
#endif

#ifndef  MACH_EPS_H
#include "Mach_eps.h" 
#endif

#ifndef VECT_H
#include "vect.h" 
#endif

#ifndef ATOM_CACHE_H
#include "Atom_cache.h" 
#endif

#ifndef INCLUDED_MATH_H
#include <math.h> 
#define INCLUDED_MATH_H
#endif

namespace MM
{

inline double
distance (Atom_cache const point_1, Atom_cache const point_2)
{
    double dx = point_1.x___ - point_2.x___;
    double dy = point_1.y___ - point_2.y___;
    double dz = point_1.z___ - point_2.z___;

    return sqrt (dx * dx + dy * dy + dz * dz);
}

inline double
distance_2 (Atom_cache const point_1, Atom_cache const point_2)
{
    double dx = point_1.x___ - point_2.x___;
    double dy = point_1.y___ - point_2.y___;
    double dz = point_1.z___ - point_2.z___;

    return dx * dx + dy * dy + dz * dz;
}

template <typename T> inline double 

length (const T & point_1,
        const T & point_2)  //fix to vector operations //fix speed
{
    double dx = point_1.x() - point_2.x();
    double dy = point_1.y() - point_2.y();
    double dz = point_1.z() - point_2.z();

    return sqrt (dx * dx + dy * dy + dz * dz);
}

template <typename T> inline double 
length_2 (const T & point_1,
        const T & point_2)  //fix to vector operations //fix speed
{
    double dx = point_1.x() - point_2.x();
    double dy = point_1.y() - point_2.y();
    double dz = point_1.z() - point_2.z();

    return dx * dx + dy * dy + dz * dz;
}

template <typename T> inline double 
angle (const T & point_1, 
       const T & point_2, 
       const T & point_3)  //fix to vector operations //fix speed
{
    double vector_21_x = point_1.x() - point_2.x();
    double vector_21_y = point_1.y() - point_2.y();
    double vector_21_z = point_1.z() - point_2.z();

    double vector_23_x = point_3.x() - point_2.x();
    double vector_23_y = point_3.y() - point_2.y();
    double vector_23_z = point_3.z() - point_2.z();

    double module_21 = sqrt (vector_21_x * vector_21_x
                           + vector_21_y * vector_21_y
                           + vector_21_z * vector_21_z);

    double module_23 = sqrt (vector_23_x * vector_23_x
                           + vector_23_y * vector_23_y
                           + vector_23_z * vector_23_z);

    //fix to k=1/mod
    vector_21_x /= module_21;
    vector_21_y /= module_21;
    vector_21_z /= module_21;

    vector_23_x /= module_23;
    vector_23_y /= module_23;
    vector_23_z /= module_23;

    double dot_product = vector_21_x * vector_23_x
                       + vector_21_y * vector_23_y
                       + vector_21_z * vector_23_z;

    return acos (dot_product);
}

template <typename T> inline double 
torsion (const T & point_a,
         const T & point_b,
         const T & point_c,
         const T & point_d)
{
    //fix Point_3D ?
    Vector a; a.x_ = point_a.x(); a.y_ = point_a.y(); a.z_ = point_a.z();  
    Vector b; b.x_ = point_b.x(); b.y_ = point_b.y(); b.z_ = point_b.z();  
    Vector c; c.x_ = point_c.x(); c.y_ = point_c.y(); c.z_ = point_c.z();  
    Vector d; d.x_ = point_d.x(); d.y_ = point_d.y(); d.z_ = point_d.z();  
    
    Vector ab = b - a;
    Vector bc = c - b;
    Vector cd = d - c;
    Vector n1 = norm (ab * bc);
    Vector n2 = norm (bc * cd);

    real  angle = (real)(n1 ^ n2);

    if (angle < -1.0)
    {
        if (angle + 1.0 < -d_mach_eps_2)
            MM::to_user().error
                ("real Model::get_torsion\n angle + 1.0 < -d_mach_eps_2");
        return  -1;
    }

    return  (((n1 * n2 ^ bc) > 0.0) ? 1 : -1) * acos (angle);
}

}//MM

#endif//GEOMETRY_H

