#ifndef COMMON_INTERACTIONS_H
#define COMMON_INTERACTIONS_H

#ifndef INTERACTION_FRAME_H
#include "Interaction_frame.h"
#endif

#ifndef FORCE_FIELD_H
#include "Force_field.h"
#endif

namespace MM
{
class Atom_group;
class Bond_group;

class Common_interactions : public Interaction_frame
{
protected:
    Force_field &           force_field_;

    double                  Van_der_Waals_potential_;
    double                  electrostatic_potential_;
    double                  bond_potential_;
    double                  angle_potential_;
    double                  torsion_potential_;
    double                  improper_torsion_potential_;
    double                  hydrophobic_potential_;

public:
    explicit Common_interactions (Force_field & ff) 
    : 
        force_field_                (ff),
        Van_der_Waals_potential_    (0.),
        electrostatic_potential_    (0.),
        bond_potential_             (0.),
        angle_potential_            (0.),
        torsion_potential_          (0.),
        improper_torsion_potential_ (0.),
        hydrophobic_potential_      (0.)
    {
    }

    virtual double      last_Van_der_Waals_potential    () const                {return Van_der_Waals_potential_;}
    virtual double      last_electrostatic_potential    () const                {return electrostatic_potential_;}
    virtual double      last_bond_potential             () const                {return bond_potential_;}
    virtual double      last_angle_potential            () const                {return angle_potential_;}
    virtual double      last_torsion_potential          () const                {return torsion_potential_;}
    virtual double      last_improper_torsion_potential () const                {return improper_torsion_potential_;}
    virtual double      last_hydrophobic_potential      () const                {return hydrophobic_potential_;}

    // Defined in subclasses
    static Common_interactions *    create (Atom_group &  atom_group,
                                            Bond_group &  bond_group,
                                            Force_field & ff);
};

}//MM

#endif //COMMON_INTERACTIONS_H
