#ifdef MM_WIN32
#pragma warning(disable : 4786)
#endif

#include "Conventional_interactions.h"

#include "Create.h"
#include "User.h"
#include "Geometry.h"

#include <math.h>
#include "Common_interactions.h"

namespace MM
{
//==============================================================================
Common_interactions * Common_interactions::
create (Atom_group & atom_group, Bond_group & bond_group, Force_field & ff)
{
    return new Conventional_interactions (atom_group, bond_group, ff);
}
//==============================================================================

double Conventional_interactions::
potential ()
{
    int i, count;
    last_potential_  = 0.;

    Van_der_Waals_potential_    = 0.;
    electrostatic_potential_    = 0.;
    bond_potential_             = 0.;
    angle_potential_            = 0.;
    torsion_potential_          = 0.;
    improper_torsion_potential_ = 0.;
    hydrophobic_potential_      = 0.;

    if (!cache_flag_)
    {
        cache_atoms ();
        cache_bonds ();
    }

    try
    {
        count = atom_count();
        for (i=0;  i<count;  ++i)
        {
            Atom_cache & current_atom = atom_[i];
            atom_potential (current_atom);
        }

        if (&bond_group_ != &nil<Bond_group>())
        {
            count = bond_count();
            for (i=0;  i<count;  ++i)
            {
                Bond_cache & current_bond = bond_[i];             
                bond_potential (current_bond);
            }
        }
    }
    catch(...)
    {
        to_user().fatal_error ("Common_interactions: Unknown exception.");
    }

    if (!cache_flag_)
    {
        atom_.clear ();
        bond_.clear ();
    }

    return last_potential_;
}

void Conventional_interactions::
zero_forces ()
{
    int count = atom_count ();

    for (int i=0;  i<count;  ++i)   
    {
        Atom_cache & atom = atom_[i];
        atom.force_x___ = 0.;
        atom.force_y___ = 0.;
        atom.force_z___ = 0.;
    }
}

void Conventional_interactions::
add_force ()
{
    int i, count;

    if (!cache_flag_)
    {
        cache_atoms ();
        cache_bonds ();
    }

    try
    {
        count = atom_count ();

        zero_forces ();      //fix remove?

        for (i=0;  i<count;  ++i)
            add_atom_force (atom_[i]);

        if (&bond_group_ != &nil<Bond_group>())
        {
            count = bond_count ();

            for (i=0;  i<count;  ++i)
                add_bond_force (bond_[i]);
        }
    }
    catch(...)
    {
        to_user().fatal_error ("Common_interactions: Unknown exception.");
    }

    if (!cache_flag_)
    {
        flush_force ();
        atom_.clear ();
        bond_.clear ();
    }
}//*/

/*void Nonbonded_interactions::
potential (Atom_cache const & atom)
{
    //__asm nop

    double  lIepsilon = 1. / epsilon_;
    double  A = 0., B = 0.;

    int     atom_number = atom.number___;
    int     count       = atom_count();

    double  charge_332  = atom.charge___ * EPS0; //331.8;
    double  e           = 0;
    double  x, y, z;
    double  lIr_6, lIr_eps;

    Force_field::VdWaals * vdw = force_field_.VdW_par [atom.mm_type___];

    double  ax = atom.x___;
    double  ay = atom.y___;
    double  az = atom.z___;

    miss_neighbours (atom);

    for (int i=atom_number+1;  i<count;  ++i)
    {
        Atom_cache & current_atom = atom_[i];

        Force_field::VdWaals & tmp = vdw [current_atom.mm_type___];  //!!!
        A = tmp.A;
        B = tmp.B;    

        x = ax - current_atom.x___;
        y = ay - current_atom.y___;
        z = az - current_atom.z___;

        lIr_6 = 1.0 / (x*x + y*y + z*z);
        lIr_eps = sqrt (lIr_6) * lIepsilon;
        lIr_6 *= lIr_6 * lIr_6;

        e += current_atom.transparency___ * (
            (A * lIr_6 - B ) * lIr_6 
            + current_atom.charge___ * charge_332 * lIr_eps);
    }

    process_14 (atom, false);
    restore_neighbours ();

    last_potential_ += e;
}//*/

void Conventional_interactions::
add_atom_force (Atom_cache & atom)
{
    add_nonbonded_force (atom);
    add_angle_force     (atom);
}

void Conventional_interactions::
add_nonbonded_force (Atom_cache & atom)
{
    double  lIepsilon = 1. / epsilon_;
    double  A = 0., B = 0.;

    int     atom_number = atom.number___;
    int     count       = atom_count();

    double  charge_332  = atom.charge___ * EPS0; //331.8;
    double  x, y, z;
    double  lIr_2, lIr_6, lIr_eps;

    Force_field::VdWaals * vdw = force_field_.VdW_par [atom.mm_type___];

    double  ax = atom.x___;
    double  ay = atom.y___;
    double  az = atom.z___;

    miss_neighbours (atom);

    for (int i=atom_number+1;  i<count;  ++i)
    {
        Atom_cache & current_atom = atom_[i];

        Force_field::VdWaals & tmp = vdw [current_atom.mm_type___];
        A = tmp.A;
        B = tmp.B;    

        x = ax - current_atom.x___;
        y = ay - current_atom.y___;
        z = az - current_atom.z___;

        lIr_2 = 1.0 / (x*x + y*y + z*z);
        lIr_eps = sqrt (lIr_2) * lIepsilon;
        lIr_6 = lIr_2 * lIr_2 * lIr_2;

        double q = current_atom.charge___ * charge_332 * lIr_eps;
        double K = - current_atom.transparency___ * 
            (6 * (B - 2 * A * lIr_6) * lIr_6 - q) * lIr_2;

        double fx = K * x;
        double fy = K * y;
        double fz = K * z;

        atom.        force_x___ += fx;
        current_atom.force_x___ -= fx;

        atom.        force_y___ += fy;
        current_atom.force_y___ -= fy;

        atom.        force_z___ += fz;
        current_atom.force_z___ -= fz;
    }
    process_14 (atom, true);
    restore_neighbours ();
}//*/

void Conventional_interactions::
add_angle_force (Atom_cache & atom_j)
{
    double  K=0,  angleEq=0;
    int     number_of_bond = atom_j.bound_count();
    int     i, k;

    for (i=0;  i<number_of_bond;  ++i)
    {
        Atom_cache & atom_i = atom_j.bound (i);

        for (k=i+1;  k<number_of_bond;  ++k)
        {
            Atom_cache & atom_k = atom_j.bound (k);

            force_field_.angle (atom_i.mm_type___, 
                                atom_j.mm_type___, 
                                atom_k.mm_type___, K, angleEq );

            double  angle_minus_angleEq = 
                MM::angle (atom_i, atom_j, atom_k) -  angleEq;

            double dV_I_dTheta = 2. * K * angle_minus_angleEq;

            //-----------------------------------------------------------------
            Vector_3D_impl Ri  (atom_i.x___, atom_i.y___, atom_i.z___);
            Vector_3D_impl Rj  (atom_j.x___, atom_j.y___, atom_j.z___);
            Vector_3D_impl Rk  (atom_k.x___, atom_k.y___, atom_k.z___);
            Vector_3D_impl Rij (Rj);    Rij -= Ri; 
            Vector_3D_impl Rkj (Rj);    Rkj -= Rk;
            Vector_3D_impl M   (Rij);   M.cross (Rkj);

            Vector_3D_impl Fi  (Rij);   Fi.cross (M);   Fi.norm ();
            Fi *=   dV_I_dTheta / Rij.length();

            Vector_3D_impl Fk  (Rkj);   Fk.cross (M);   Fk.norm ();
            Fk *= - dV_I_dTheta / Rkj.length();

            Vector_3D_impl Fj  (Fi);    Fj += Fk;       Fj.negate ();

            atom_i.force_x___ += Fi.x();
            atom_i.force_y___ += Fi.y();
            atom_i.force_z___ += Fi.z();

            atom_j.force_x___ += Fj.x();
            atom_j.force_y___ += Fj.y();
            atom_j.force_z___ += Fj.z();

            atom_k.force_x___ += Fk.x();
            atom_k.force_y___ += Fk.y();
            atom_k.force_z___ += Fk.z();
        }
    }
}

void Conventional_interactions::
add_bond_force (Bond_cache & bond)
{
    add_bond_stretch_force (bond);
    add_torsion_force      (bond);
}

void Conventional_interactions::
add_bond_stretch_force (Bond_cache & bond)
{
    // Harmonic Bond =========================================================
    double Kr, Req;

    Atom_cache & atom_1 = *bond.first_atom___;
    Atom_cache & atom_2 = *bond.second_atom___;

    force_field_.bond (atom_1.mm_type___, atom_2.mm_type___,  Kr, Req );

    double R           = distance (atom_1, atom_2);
    double R_minus_Req = R - Req;
    double dPot_I_dR   = 2. * Kr * R_minus_Req;

    double dPot_I_dR_I_R =  dPot_I_dR / R;
    double fx = (atom_2.x___ - atom_1.x___) * dPot_I_dR_I_R;
    double fy = (atom_2.y___ - atom_1.y___) * dPot_I_dR_I_R;
    double fz = (atom_2.z___ - atom_1.z___) * dPot_I_dR_I_R;;

    atom_1.force_x___ += fx;
    atom_2.force_x___ -= fx;

    atom_1.force_y___ += fy;
    atom_2.force_y___ -= fy;

    atom_1.force_z___ += fz;
    atom_2.force_z___ -= fz;
}

void Conventional_interactions::
add_torsion_force (Bond_cache & bond_j)
{
    Atom_cache & atom_j = *bond_j.first_atom___;
    Atom_cache & atom_k = *bond_j.second_atom___;

    int     i,l, tor;

    for (i=0;  i<atom_j.bound_count();  ++i)
    {
        //Atom_cache const & atom_1 = atom_j.bound (i);
        Atom_cache & atom_i = atom_j.bound (i);
        //if (&atom_1 == &atom_3) continue;
        if (&atom_i == &atom_k) continue;

        //for (j=0;  j<atom_3.bound_count();  j++)
        for (l=0;  l<atom_k.bound_count();  ++l)
        {
            //Atom_cache const & atom_4 = atom_3.bound (j);
            Atom_cache & atom_l = atom_k.bound (l);
            //if (&atom_j == &atom_4) continue;
            if (&atom_j == &atom_l) continue;
                                                           //fix may be 0             
            const Array <Force_field::Torsions> * tor_tmp; //!!!
            tor_tmp = force_field_.torsion (atom_i.mm_type___, 
                                            atom_j.mm_type___, 
                                            atom_k.mm_type___, 
                                            atom_l.mm_type___);

            //k_max = tor_tmp->size();
            //for (k=0;  k<k_max;  k++)
            int tor_max = tor_tmp->size();
            for (tor=0;  tor<tor_max;  tor++)
            {
                int    Mult  = (*tor_tmp)[tor].Mult;
                double V_2   = (*tor_tmp)[tor].half_V;

                if (V_2 < d_mach_eps_2)
                    continue;

                double Phi0 = (*tor_tmp)[tor].Fi0;
                int    n    = (*tor_tmp)[tor].N;

                double Phi = MM::torsion (atom_i, atom_j, atom_k, atom_l); 

                // e += V_2 * (1 + cos (n*Fi - Fi0)) / Mult;
                double dV_I_dPhi = - V_2 * n * sin (n * Phi - Phi0) / Mult;

                Vector_3D_impl  Ri  (atom_i.x___, atom_i.y___, atom_i.z___);
                Vector_3D_impl  Rj  (atom_j.x___, atom_j.y___, atom_j.z___);
                Vector_3D_impl  Rk  (atom_k.x___, atom_k.y___, atom_k.z___);
                Vector_3D_impl  Rl  (atom_l.x___, atom_l.y___, atom_l.z___);
                Vector_3D_impl  Rij (Rj);   Rij -= Ri;
                Vector_3D_impl  Rkj (Rj);   Rkj -= Rk;
                Vector_3D_impl  Rkl (Rl);   Rkl -= Rk;
                Vector_3D_impl  M   (Rij);  M.cross (Rkj);
                Vector_3D_impl  N   (Rkj);  N.cross (Rkl);
                double          kj_length_2 = Rkj.length_2 ();
                double          kj_length   = sqrt (kj_length_2);

                Vector_3D_impl Fi  (M);
                Fi *= - dV_I_dPhi * kj_length / M.length_2();

                Vector_3D_impl Fl  (N);
                Fl *=   dV_I_dPhi * kj_length / N.length_2();

                atom_i.force_x___ += Fi.x(); 
                atom_i.force_y___ += Fi.y(); 
                atom_i.force_z___ += Fi.z();

                atom_l.force_x___ += Fl.x(); 
                atom_l.force_y___ += Fl.y(); 
                atom_l.force_z___ += Fl.z();

                double Rij_dot_Rkj = Rij.dot(Rkj);  Rij_dot_Rkj /= kj_length_2;
                double Rkl_dot_Rkj = Rkl.dot(Rkj);  Rkl_dot_Rkj /= kj_length_2;

                Vector_3D_impl Fj  (Fi);    Fj.negate ();
                Vector_3D_impl Fk  (Fl);    Fk.negate ();

                Fi *= Rij_dot_Rkj;
                Fl *= Rkl_dot_Rkj;

                Fj += Fi;   Fj -= Fl;
                Fk -= Fi;   Fk += Fl;

                atom_j.force_x___ += Fj.x(); 
                atom_j.force_y___ += Fj.y(); 
                atom_j.force_z___ += Fj.z();

                atom_k.force_x___ += Fk.x(); 
                atom_k.force_y___ += Fk.y(); 
                atom_k.force_z___ += Fk.z();

                ////    Fi = getTorsion( A1, A2, A3, A4 );
                //double  ab_x, ab_y, ab_z;
                //ab_x= atom_2.x___ -  atom_1.x___;
                //ab_y= atom_2.y___ -  atom_1.y___;
                //ab_z= atom_2.z___ -  atom_1.z___;
                //double  bc_x, bc_y, bc_z;
                //bc_x= atom_3.x___ -  atom_j.x___;
                //bc_y= atom_3.y___ -  atom_j.y___;
                //bc_z= atom_3.z___ -  atom_j.z___;
                //double  cd_x, cd_y, cd_z;
                //cd_x= atom_4.x___ -  atom_3.x___;
                //cd_y= atom_4.y___ -  atom_3.y___;
                //cd_z= atom_4.z___ -  atom_3.z___;

                ////    TVector  n1 = (ab*bc)();
                ////    TVector  n2 = (bc*cd)();
                //double  Xn1, Yn1, Zn1,  Xn2, Yn2, Zn2,  Norm;
                //Xn1 = ab_y * bc_z - ab_z * bc_y;
                //Yn1 = ab_z * bc_x - ab_x * bc_z;
                //Zn1 = ab_x * bc_y - ab_y * bc_x;

                //Xn2 = bc_y * cd_z - bc_z * cd_y;
                //Yn2 = bc_z * cd_x - bc_x * cd_z;
                //Zn2 = bc_x * cd_y - bc_y * cd_x;

                //Norm = sqrt ((Xn1*Xn1+Yn1*Yn1+Zn1*Zn1)*
                //             (Xn2*Xn2+Yn2*Yn2+Zn2*Zn2));

                ////    real cos_nFi = n1^n2;
                //double cos_nFi = (Xn1*Xn2+Yn1*Yn2+Zn1*Zn2) / Norm;

                //switch (n)                                    //fix  for Fi0 0 or 180 only
                //{
                //    case 1:   e += (Fi0 == 0) ?
                //                    V_2 * (1 + cos_nFi) / Mult :
                //                    V_2 * (1 - cos_nFi) / Mult;
                //              break;

                //    case 2:   e += (Fi0 == 0) ?
                //                    V_2 * (2*cos_nFi*cos_nFi) / Mult :
                //                    V_2 * 2 * (1 - cos_nFi*cos_nFi) / Mult;
                //              break;

                //    case 3:   e += (Fi0 == 0) ?
                //                    V_2 * (1 + (4*cos_nFi*cos_nFi*cos_nFi -
                //                    3 * cos_nFi)) / Mult :
                //                    V_2 * (1 - (4*cos_nFi*cos_nFi*cos_nFi -
                //                    3 * cos_nFi)) / Mult;
                //              break;

                //    default:  real  Fi = MM::torsion
                //                (atom_1, atom_2, atom_3, atom_4);

                //              e += V_2 * (1 + cos (n*Fi - Fi0)) / Mult;
                //}
            }
        }                
    }
}

void Conventional_interactions::
atom_potential (Atom_cache & atom)
{
    nonbonded_potential  (atom);
    angle_potential (atom);
}

void Conventional_interactions::
bond_potential (Bond_cache const & bond)
{
    bond_stretch_potential (bond);
    torsion_potential (bond);
}

void Conventional_interactions::
nonbonded_potential (Atom_cache & atom)
{
    // Nonbonded interactions =================================================

    double  lIepsilon = 1. / epsilon_;
    double  A = 0., B = 0.;

    int     atom_number = atom.number___;
    int     count       = atom_count();

    double  charge_332  = atom.charge___ * EPS0; //331.8;
    //double  e           = 0;
    double  P_VdW = 0., P_el = 0.;
    double  x, y, z;
    double  lIr_6, lIr_eps;

    Force_field::VdWaals * vdw = force_field_.VdW_par [atom.mm_type___];

    double  ax = atom.x___;
    double  ay = atom.y___;
    double  az = atom.z___;

    miss_neighbours (atom);

    // 0.00315
    for (int i=atom_number+1;  i<count;  i+=4)
    {
        Atom_cache & current_atom  = atom_[i];
        Atom_cache & current_atom2 = atom_[i+1];
        Atom_cache & current_atom3 = atom_[i+2];
        Atom_cache & current_atom4 = atom_[i+3];

        Force_field::VdWaals & tmp = vdw [current_atom.mm_type___];  //!!!
        A = tmp.A;
        B = tmp.B; 

        x = ax - current_atom.x___;
        y = ay - current_atom.y___;
        z = az - current_atom.z___;

        lIr_6 = 1.0 / (x*x + y*y + z*z);
        lIr_eps = sqrt (lIr_6) * lIepsilon;
        lIr_6 *= lIr_6 * lIr_6;

        //e += current_atom.transparency___ * (
        //    (A * lIr_6 - B ) * lIr_6 
        //    + current_atom.charge___ * charge_332 * lIr_eps);
        P_VdW += current_atom.transparency___ * 
            (A * lIr_6 - B ) * lIr_6;

        P_el += current_atom.transparency___ *
            current_atom.charge___ * charge_332 * lIr_eps;

        Force_field::VdWaals & tmp2 = vdw [current_atom2.mm_type___];  //!!!
        A = tmp2.A;
        B = tmp2.B;    

        x = ax - current_atom2.x___;
        y = ay - current_atom2.y___;
        z = az - current_atom2.z___;

        lIr_6 = 1.0 / (x*x + y*y + z*z);
        lIr_eps = sqrt (lIr_6) * lIepsilon;
        lIr_6 *= lIr_6 * lIr_6;

        //e += current_atom2.transparency___ * (
        //    (A * lIr_6 - B ) * lIr_6 
        //    + current_atom2.charge___ * charge_332 * lIr_eps);
        P_VdW += current_atom2.transparency___ * 
            (A * lIr_6 - B ) * lIr_6;

        P_el += current_atom2.transparency___ *
            current_atom2.charge___ * charge_332 * lIr_eps;


        Force_field::VdWaals & tmp3 = vdw [current_atom3.mm_type___];  //!!!
        A = tmp3.A;
        B = tmp3.B;    

        x = ax - current_atom3.x___;
        y = ay - current_atom3.y___;
        z = az - current_atom3.z___;

        lIr_6 = 1.0 / (x*x + y*y + z*z);
        lIr_eps = sqrt (lIr_6) * lIepsilon;
        lIr_6 *= lIr_6 * lIr_6;

        //e += current_atom3.transparency___ * (
        //    (A * lIr_6 - B ) * lIr_6 
        //    + current_atom3.charge___ * charge_332 * lIr_eps);
        P_VdW += current_atom3.transparency___ * 
            (A * lIr_6 - B ) * lIr_6;

        P_el += current_atom3.transparency___ *
            current_atom3.charge___ * charge_332 * lIr_eps;

        Force_field::VdWaals & tmp4 = vdw [current_atom4.mm_type___];  //!!!
        A = tmp4.A;
        B = tmp4.B;    

        x = ax - current_atom4.x___;
        y = ay - current_atom4.y___;
        z = az - current_atom4.z___;

        lIr_6 = 1.0 / (x*x + y*y + z*z);
        lIr_eps = sqrt (lIr_6) * lIepsilon;
        lIr_6 *= lIr_6 * lIr_6;

        //e += current_atom4.transparency___ * (
        //    (A * lIr_6 - B ) * lIr_6 
        //    + current_atom4.charge___ * charge_332 * lIr_eps);
        P_VdW += current_atom4.transparency___ * 
            (A * lIr_6 - B ) * lIr_6;

        P_el += current_atom4.transparency___ *
            current_atom4.charge___ * charge_332 * lIr_eps;
    }

    process_14 (atom, false);
    restore_neighbours ();

    Van_der_Waals_potential_ += P_VdW;
    electrostatic_potential_ += P_el;
    last_potential_ += P_el + P_VdW;
}

void Conventional_interactions::
angle_potential (Atom_cache const & atom_2)
{
    double  e = 0,  K=0,  angleEq=0;

    int     number_of_bond = atom_2.bound_count();
    int     i, k;

    for (i=0;  i<number_of_bond;  i++)
    {
        Atom_cache const & atom_1 = atom_2.bound (i);

        for (k=i+1;  k<number_of_bond;  k++)
        {
            Atom_cache const & atom_3 = atom_2.bound (k);
            force_field_.angle (atom_1.mm_type___, 
                                atom_2.mm_type___, 
                                atom_3.mm_type___, 
                                K, angleEq );

            double  angle_minus_angleEq = 
                MM::angle (atom_1, atom_2, atom_3) -  angleEq;

            e += K * angle_minus_angleEq * angle_minus_angleEq;
        }
    }
    angle_potential_ += e;
    last_potential_  += e; 
}

void Conventional_interactions::
bond_stretch_potential (Bond_cache const & bond)
{
    // Harmonic Bond =========================================================
    double Kr, Req;

    Atom_cache const & atom_1 = *bond.first_atom___;
    Atom_cache const & atom_2 = *bond.second_atom___;

    force_field_.bond (atom_1.mm_type___, atom_2.mm_type___,  Kr, Req );

    double  R_minus_Req = distance (atom_1, atom_2) - Req;

    double e = Kr * R_minus_Req * R_minus_Req;
    
    bond_potential_ += e;
    last_potential_ += e;
}

void Conventional_interactions::
torsion_potential (Bond_cache const & bond)
{
    // e = V/2 ( 1 + cos( n*Fi - Fi0 ) )
    double e = 0;
    const Atom_cache & atom_2 = *bond.first_atom___;
    const Atom_cache & atom_3 = *bond.second_atom___;

    int     i,j,k, k_max;

    for (i=0;  i<atom_2.bound_count();  i++)
    {
        Atom_cache const & atom_1 = atom_2.bound (i);
        if (&atom_1 == &atom_3) continue;

        for (j=0;  j<atom_3.bound_count();  j++)
        {
            Atom_cache const & atom_4 = atom_3.bound (j);
            if (&atom_2 == &atom_4) continue;
                                                           //fix may be 0             
            const Array <Force_field::Torsions> * tor_tmp; //!!!
            tor_tmp = force_field_.torsion (atom_1.mm_type___, 
                                            atom_2.mm_type___, 
                                            atom_3.mm_type___, 
                                            atom_4.mm_type___);

            k_max = tor_tmp->size();
            for (k=0;  k<k_max;  k++)
            {
                int    Mult  = (*tor_tmp)[k].Mult;
                double V_2   = (*tor_tmp)[k].half_V;

                if (V_2 < d_mach_eps_2)
                    continue;

                double Fi0   = (*tor_tmp)[k].Fi0;
                int    n     = (*tor_tmp)[k].N;

                //    Fi = getTorsion( A1, A2, A3, A4 );
                double  ab_x, ab_y, ab_z;
                ab_x= atom_2.x___ -  atom_1.x___;
                ab_y= atom_2.y___ -  atom_1.y___;
                ab_z= atom_2.z___ -  atom_1.z___;
                double  bc_x, bc_y, bc_z;
                bc_x= atom_3.x___ -  atom_2.x___;
                bc_y= atom_3.y___ -  atom_2.y___;
                bc_z= atom_3.z___ -  atom_2.z___;
                double  cd_x, cd_y, cd_z;
                cd_x= atom_4.x___ -  atom_3.x___;
                cd_y= atom_4.y___ -  atom_3.y___;
                cd_z= atom_4.z___ -  atom_3.z___;

                //    TVector  n1 = (ab*bc)();
                //    TVector  n2 = (bc*cd)();
                double  Xn1, Yn1, Zn1,  Xn2, Yn2, Zn2,  Norm;
                Xn1 = ab_y * bc_z - ab_z * bc_y;
                Yn1 = ab_z * bc_x - ab_x * bc_z;
                Zn1 = ab_x * bc_y - ab_y * bc_x;

                Xn2 = bc_y * cd_z - bc_z * cd_y;
                Yn2 = bc_z * cd_x - bc_x * cd_z;
                Zn2 = bc_x * cd_y - bc_y * cd_x;

                Norm = sqrt ((Xn1*Xn1+Yn1*Yn1+Zn1*Zn1)*
                             (Xn2*Xn2+Yn2*Yn2+Zn2*Zn2));

                //    real cos_nFi = n1^n2;
                double cos_nFi = (Xn1*Xn2+Yn1*Yn2+Zn1*Zn2) / Norm;

                switch (n)                                    //fix  for Fi0 0 or 180 only
                {
                    case 1:   e += (Fi0 == 0) ?
                                    V_2 * (1 + cos_nFi) / Mult :
                                    V_2 * (1 - cos_nFi) / Mult;
                              break;

                    case 2:   e += (Fi0 == 0) ?
                                    V_2 * (2*cos_nFi*cos_nFi) / Mult :
                                    V_2 * 2 * (1 - cos_nFi*cos_nFi) / Mult;
                              break;

                    case 3:   e += (Fi0 == 0) ?
                                    V_2 * (1 + (4*cos_nFi*cos_nFi*cos_nFi -
                                    3 * cos_nFi)) / Mult :
                                    V_2 * (1 - (4*cos_nFi*cos_nFi*cos_nFi -
                                    3 * cos_nFi)) / Mult;
                              break;

                    default:  real  Fi = MM::torsion
                                (atom_1, atom_2, atom_3, atom_4);

                              e += V_2 * (1 + cos (n*Fi - Fi0)) / Mult;
                }
            }
        }                
    }

    torsion_potential_ += e;
    last_potential_    += e;
}

/*void Nonbonded_interactions::
add_force (Atom_cache & atom)
{
    double  lIepsilon = 1. / epsilon_;
    double  A = 0., B = 0.;

    int     atom_number = atom.number___;
    int     count       = atom_count();

    double  charge_332  = atom.charge___ * EPS0; //331.8;
    double  x, y, z;
    double  lIr_2, lIr_6, lIr_eps;

    Force_field::VdWaals * vdw = force_field_.VdW_par [atom.mm_type___];

    double  ax = atom.x___;
    double  ay = atom.y___;
    double  az = atom.z___;

    miss_neighbours (atom);

    for (int i=atom_number+1;  i<count;  i+=2)
    {
        Atom_cache & current_atom  = atom_[i];
        Atom_cache & current_atom2 = atom_[i+1];

        Force_field::VdWaals & tmp = vdw [current_atom.mm_type___];
        A = tmp.A;
        B = tmp.B;    

        x = ax - current_atom.x___;
        y = ay - current_atom.y___;
        z = az - current_atom.z___;

        lIr_2 = 1.0 / (x*x + y*y + z*z);
        lIr_eps = sqrt (lIr_2) * lIepsilon;
        lIr_6 = lIr_2 * lIr_2 * lIr_2;

        double q = current_atom.charge___ * charge_332 * lIr_eps;
        double K = - current_atom.transparency___ * 
            (6 * (B - 2 * A * lIr_6) * lIr_6 - q) * lIr_2;

        double fx = K * x;
        double fy = K * y;
        double fz = K * z;

        atom.        force_x___ += fx;
        current_atom.force_x___ -= fx;

        atom.        force_y___ += fy;
        current_atom.force_y___ -= fy;

        atom.        force_z___ += fz;
        current_atom.force_z___ -= fz;


        Force_field::VdWaals & tmp2 = vdw [current_atom2.mm_type___];
        A = tmp2.A;
        B = tmp2.B;    

        x = ax - current_atom2.x___;
        y = ay - current_atom2.y___;
        z = az - current_atom2.z___;

        lIr_2 = 1.0 / (x*x + y*y + z*z);
        lIr_eps = sqrt (lIr_2) * lIepsilon;
        lIr_6 = lIr_2 * lIr_2 * lIr_2;

        q = current_atom2.charge___ * charge_332 * lIr_eps;
        K = - current_atom2.transparency___ * 
            (6 * (B - 2 * A * lIr_6) * lIr_6 - q) * lIr_2;

        fx = K * x;
        fy = K * y;
        fz = K * z;

        atom.         force_x___ += fx;
        current_atom2.force_x___ -= fx;

        atom.         force_y___ += fy;
        current_atom2.force_y___ -= fy;

        atom.         force_z___ += fz;
        current_atom2.force_z___ -= fz;
    }

    restore_neighbours ();
}//*/

void Conventional_interactions::
process_14 (Atom_cache & atom_1, bool force)
{
    for (int i=0;  i<atom_1.bound_count();  i++)
    {
        Atom_cache & atom_2 = atom_1.bound (i);

        for (int j=0;  j<atom_2.bound_count();  j++)
        {
            Atom_cache & atom_3 = atom_2.bound (j);
            if (&atom_1 == &atom_3) continue;

            for (int k=0;  k<atom_3.bound_count();  k++)
            {
                Atom_cache & atom_4 = atom_3.bound (k);
                
                if (&atom_2 == &atom_4 || 
                     atom_4.flag___.is (Atom_kit::missed)) 
                        continue;
                
                atom_4.flag___.set (Atom_kit::missed);
                miss_set_.insert (&atom_4);

                if (atom_1.number___ >= atom_4.number___)
                    continue;

                if (force)
                    add_nonbonded_force_14 (atom_1, atom_4);
                else
                    nonbonded_potential_14 (atom_1, atom_4);
            }
        }
    }
}

void Conventional_interactions::
nonbonded_potential_14 (Atom_cache & atom_1, Atom_cache & atom_2)
{
//    double  e = 0;
    double  P_VdW = 0., P_el = 0.;
    double  A, B;

    // e = A/r^12 - B/r^6  +  electrostatics
    force_field_.VdW (atom_1.mm_type___, atom_2.mm_type___, A, B);

    double  x = atom_1.x___ - atom_2.x___;
    double  y = atom_1.y___ - atom_2.y___;
    double  z = atom_1.z___ - atom_2.z___;

    double lIr_2 = 1. / (x*x + y*y + z*z);            // 1 / r^2
    double lIr_eps = sqrt (lIr_2) / force_field_.eps; // 1 / r * eps
    double lIr_6 = lIr_2 * lIr_2 * lIr_2;             // 1 / r^6

    P_VdW += (A * lIr_6 - B) * lIr_6 * force_field_.VdW_Scale;
    P_el += atom_1.charge___ * atom_2.charge___ * EPS0 * 
        force_field_.E_Static_Scale * lIr_eps;

    Van_der_Waals_potential_ += P_VdW;
    electrostatic_potential_ += P_el;
    last_potential_          += P_VdW + P_el;
}

void Conventional_interactions::
add_nonbonded_force_14 (Atom_cache & atom_1, Atom_cache & atom_2)
{
    double  A, B;

    // e = A/r^12 - B/r^6  +  electrostatics
    force_field_.VdW (atom_1.mm_type___, atom_2.mm_type___, A, B);

    double  x = atom_1.x___ - atom_2.x___;
    double  y = atom_1.y___ - atom_2.y___;
    double  z = atom_1.z___ - atom_2.z___;

    double lIr_2 = 1. / (x*x + y*y + z*z);            // 1 / r^2
    double lIr_eps = sqrt (lIr_2) / force_field_.eps; // 1 / r * eps
    double lIr_6 = lIr_2 * lIr_2 * lIr_2;             // 1 / r^6

    double K  = - force_field_.VdW_Scale * 6 * (B - 2 * A * lIr_6) * 
        (lIr_6*lIr_2) - atom_1.charge___ * atom_2.charge___ * EPS0 * 
        force_field_.E_Static_Scale * (lIr_eps * lIr_2);

    double fx = K * x;
    double fy = K * y;
    double fz = K * z;

    atom_1.force_x___ += fx;
    atom_2.force_x___ -= fx;

    atom_1.force_y___ += fy;
    atom_2.force_y___ -= fy;

    atom_1.force_z___ += fz;
    atom_2.force_z___ -= fz;
}

void Conventional_interactions::
miss_neighbours (Atom_cache const & atom_1)
{
    REQUIRE ("", miss_set_.size() == 0);
    int i, j, k;

    for (i=0;  i<atom_1.bound_count();  i++)
    {
        Atom_cache const & atom_2 = atom_1.bound (i);

        if (atom_2.flag___.is_not (Atom_kit::missed))
        {
            atom_2.transparency___ = 0.;
            atom_2.flag___.set (Atom_kit::missed);
            miss_set_.insert (&atom_2);
        }

        for (j=0;  j<atom_2.bound_count();  j++)
        {
            Atom_cache const & atom_3 = atom_2.bound (j);
            if (&atom_1 == &atom_3) continue;

            if (atom_3.flag___.is_not (MM::Atom_kit::missed))
            {
                atom_3.transparency___ = 0.;
                atom_3.flag___.set (Atom_kit::missed);
                miss_set_.insert (&atom_3);
            }
            
            for (k=0;  k<atom_3.bound_count();  k++)
            {
                Atom_cache const & atom_4 = atom_3.bound (k);

                if (&atom_2 == &atom_4) continue;

                if (atom_4.flag___.is_not (Atom_kit::missed))
                    atom_4.transparency___ = 0.;
            }
        }
    }
}

void Conventional_interactions::
restore_neighbours ()
{
    std::set <Atom_cache const *>::iterator it = miss_set_.begin();
    for (;  it!=miss_set_.end();  ++it)
    {
        (*it)->flag___.clear (Atom_kit::missed);
        (*it)->transparency___ = 1.;
    }
    miss_set_.clear();
}

Conventional_interactions::
Conventional_interactions (Atom_group & atom_group, Force_field & ff)
:
    Common_interactions         (ff),
    Bond_group_cache (nil<Bond_group>(), atom_group),
    epsilon_                    (1.)
{
}

Conventional_interactions::
Conventional_interactions (Atom_group & atom_group, Bond_group & bond_group,
    Force_field & ff)
:
    Common_interactions         (ff),
    Bond_group_cache (bond_group, atom_group),
    epsilon_                    (1.)
{
}

void Conventional_interactions::
report (Text *result, Text const & hint)
{
}

}//MM

