#ifndef NONBONDED_INTERACTIONS_H
#define NONBONDED_INTERACTIONS_H

#ifndef BOND_GROUP_CACHE_H
#include "Bond_group_cache.h"
#endif

#ifndef INTERACTION_FRAME_H
#include "Interaction_frame.h"
#endif

#ifndef ENERGY_H
#include "Energy.h"
#endif

#ifndef COMMON_INTERACTIONS_H
#include "Common_interactions.h"
#endif

#include <set>

namespace MM
{
struct Bond_cache;
struct Atom_cache;

class Conventional_interactions :
    public Bond_group_cache,
    public Common_interactions
{
protected:
    std::set <Atom_cache const *> miss_set_;

    double                  epsilon_;

public:
    Conventional_interactions (Atom_group & atom_group, Force_field & ff);

    Conventional_interactions (Atom_group & atom_group,
                         Bond_group & bond_group, Force_field & ff);

    virtual double      potential ();
    virtual void        add_force ();

    virtual Text        title () const {return "Conventional";}
    virtual void        report (Text *result, Text const & hint = Text("Full"));

protected:
    friend class Covalent_interactions_test;

    virtual void        zero_forces ();

    virtual int         atom_count()    {return atom_group_.atom_count();}
    virtual int         bond_count()    {return bond_group_.bond_count();}

    virtual void        atom_potential          (Atom_cache & atom);
    virtual void        bond_potential          (Bond_cache const & bond);

    virtual void        nonbonded_potential     (Atom_cache       & atom);
    virtual void        angle_potential         (Atom_cache const & atom);
    virtual void        bond_stretch_potential  (Bond_cache const & bond);
    virtual void        torsion_potential       (Bond_cache const & bond);

    virtual void        add_atom_force          (Atom_cache &       atom);
    virtual void        add_bond_force          (Bond_cache &       bond);

    virtual void        add_nonbonded_force     (Atom_cache & atom);
    virtual void        add_angle_force         (Atom_cache & atom);
    virtual void        add_bond_stretch_force  (Bond_cache & bond);
    virtual void        add_torsion_force       (Bond_cache & bond);

    virtual void        process_14              (Atom_cache &, bool force);
    virtual void        nonbonded_potential_14  (Atom_cache &, Atom_cache &);
    virtual void        add_nonbonded_force_14  (Atom_cache &, Atom_cache &);
    virtual void        miss_neighbours         (Atom_cache const &);
    virtual void        restore_neighbours      ();
};

}//MM

#endif //NONBONDED_INTERACTIONS_H
