#ifndef COLLECT_SCALAR_H
#define COLLECT_SCALAR_H

#ifndef COLLECT_QUANTITY_H
#include "Collect_quantity.h"
#endif

#ifndef ARRAY_H
#include "Array.h"
#endif

#ifndef COLLECT_SCALAR_PRESENTATION_H
#include "Collect_scalar_presentation.h"
#endif

namespace MM
{

class Collect_scalar : public Collect_quantity
{
protected:
    Array <double>      quantity_sum_;
    Array <double>      quantity_2_sum_;

public:
    explicit Collect_scalar (Atom_group  & atom_group);
    explicit Collect_scalar (Atom_group  & atom_group,
                             Collect_scalar_presentation *new_presentation);

    void                clear();

    double  value   (int i) const   
    {
        REQUIRE ("In range.", i >= 0 && i < count());

        return i == 0 
            ? quantity_sum_[0] 
            : quantity_sum_[i] - quantity_sum_[i-1];
    }

    bool    has_values () const     {return quantity_sum_.size() > 0;}
    double  last_value () const     {REQUIRE ("", has_values ());   return value   (count()-1);}
    double  average () const        {REQUIRE ("", has_values ());   return average (count()-1);}
    double  rmsd    () const        {REQUIRE ("", has_values ());   return rmsd    (count()-1);}
    double  average (int i) const   {REQUIRE ("", has_values ());   return quantity_sum_[i] / (i+1);}
    double  rmsd    (int i) const;

    Collect_scalar *    adopt (Collect_scalar_presentation *new_presentation);

protected:
    virtual double      extract_current_value () =0;

    void                    handle_execute ();
    Monitor_presentation *  create_file_presentation  (Text const& filename);
    Monitor_presentation *  create_graph_presentation (Text const& graphname);
};

}//MM

#endif //COLLECT_SCALAR_H
