#ifndef MD_H
#define MD_H

#ifndef SIMULATION_H
#include "Simulation.h"
#endif

#ifndef TEXT_H
#include "Text.h"
#endif

#ifndef MD_SUBJECT_H
#include "MD_subject.h"
#endif

#ifndef EVENT_H
#include "Event.h"
#endif

namespace MM
{
class MD_method;

class MD : public Simulation, protected MD_subject
{
protected:
    double                              step_length_;   // AKMA time units
    double                              drift_length_;  //fix Why? Sum!
    double                              kick_length_;

    MD_subject *                        subject_;
    MD_method  *                        method_;

    Event                               step_finished_;
    Event                               stop_;

public:
    MD ();
    explicit MD (MD_subject & subject, MD_method & method);

    virtual char const *    class_name () const {return "MD";}
    MD_subject &    interface_for (MD_method const*) {return *this;}

    //virtual MD *    clone (MD_subject & subject) const =0;

    // externally in picoseconds and internally in AKMA time units
    double  step_length () const;
    void    set_step_length (double new_length);

    void    set_method  (MD_method & new_method);
    void    set_subject (MD_subject & new_subject);

    void    after_each_step (Command * to_do)
    {
        step_finished_.subscribe (to_do);
    }

    void    eliminate_after_each_step (Command & to_do)
    {
        step_finished_.subscribe (&to_do, false);
    }

    void    after_stop (Command * to_do)
    {
        stop_.subscribe (to_do);
    }

    void    eliminate_after_stop (Command & to_do)
    {
        stop_.subscribe (&to_do, false);
    }

    virtual Text    title () const;
//        { static const Text name = "MD";    return name;}

protected:
    virtual void        start      ();
    virtual void        cycle      ();     //fix loop
    //virtual void        step       () =0;
    virtual int         iterations () const;
    virtual void        finish     ();

    virtual void        kick      (double h);
    virtual void        drift     (double h);

protected:
    //bool                invariant () const;
};

}//MM

#endif //MD_H
