#include "MD_simulator.h"

#include "Create.h"

#include "Defs.h"
#include "Own.h"
#include "Model.h"
#include "Model_kit.h"
#include "Project.h"
#include "Atom_group.h"
#include "Interaction.h"
#include "MD.h"
#include "MD_subject.h"
#include "Monitor.h"
#include "Atom_group_as_md_subject.h"
#include "Mass.h"
#include "Arbitrary_atom_group.h"
#include "No_interaction.h"

#include "PEFRL.h"
#include "VEFRL.h"
#include "OPVL.h"
#include "OVVL.h"
#include "Position_leapfrog.h"
#include "Velocity_leapfrog.h"
#include "Velocity_verlet.h"
#include "Position_verlet.h"
#include "Simulator_factory.h"
//#include "Verlet.h"

#include "Collect_phase_trajectory.h"
#include "Collect_energy.h"

namespace MM
{
Array_of_own <MD_method>    MD_simulator::method_prototype_;
MD_simulator MD_simulator::prototype_ (it_is_a_prototype);

MD_simulator::MD_simulator (Prototype &)
{
    Simulator_factory::accept ("MD", this);
}

MD_simulator::
MD_simulator ()
:
    //Simulator_frame (Project::singleton().current_model()),
    method_     (MD_method::create ("Velocity_verlet")),
    simulation_ (new MD),
    starting_temperature_(Constant::normal_temperature),
    starting_pressure_   (Constant::normal_pressure),
    final_temperature_   (Constant::normal_temperature),
    final_pressure_      (Constant::normal_pressure)
{
    simulation_().set_method (method_()); //fix move to run
}//*/

MD_simulator::
MD_simulator (Model &      model,
              char const * method_name)
:
    Simulator_frame (model),
    method_  (MD_method::create (method_name)),
    simulation_ (new MD),
    starting_temperature_(Constant::normal_temperature),
    starting_pressure_   (Constant::normal_pressure),
    final_temperature_   (Constant::normal_temperature),
    final_pressure_      (Constant::normal_pressure)
{
    simulation_().set_method (method_()); //fix move to run
}

void MD_simulator::
add_monitor (char const * name)
{
    Monitor * new_monitor = Monitor::create (model(), name);
    simulation_(). after_each_step (new_monitor);
    monitor_.push_back (*new_monitor);
}

void MD_simulator::
tune ()
{
    subject_.adopt (model().kit().create_md_interface ());
    simulation_().set_subject (subject_());
    clear_monitors ();
}

void MD_simulator::
start ()
{
    MD & md = simulation_();
    md.run();
}

int MD_simulator::
steps () const
{
    return simulation_().steps ();
}

void MD_simulator::
set_steps (int steps)
{
    simulation_(). set_steps (steps);
}

Duration_unit MD_simulator::   // picoseconds
step_length () const
{
//    return simulation_(). step_length() * Constant::time_unit_to_psec;
    Duration_unit result;
    result.set_ps (simulation_().step_length() * Constant::time_unit_to_psec);
    return result;
}

void MD_simulator::    // picoseconds
set_step_length (Duration_unit new_length)
{
    MD & md = simulation_();
    md. set_step_length
        (new_length.ps() * Constant::psec_to_time_unit);

    md. set_steps ((int)new_length.steps());
}

void MD_simulator::
after_each_step (Command * to_do)
{
    simulation_(). after_each_step (to_do);
}

void MD_simulator::
after_stop (Command * to_do)
{
    simulation_(). after_stop (to_do);
}

void MD_simulator::
eliminate_after_each_step (Command & to_do)
{
    simulation_(). eliminate_after_each_step (to_do);
}

void MD_simulator::
eliminate_after_stop (Command & to_do)
{
    simulation_(). eliminate_after_stop (to_do);
}
/*
void MD_simulator::
unsubscribe (MD_monitor & to_do)
{
    algorithm(). unsubscribe (to_do);
    monitor_.    remove      (to_do);
}
//*/

void MD_simulator::
update ()
{
    //fix show panel
}

}//MM
