#include "Form_add_model.h"

#include <qlineedit.h>
#include <qcheckbox.h>
#include <qpushbutton.h>
#include <qvalidator.h>

#include "Path.h"
#include "Qt_application.h"
#include "Defs.h"
#include "Model.h"
#include "Project.h"
#include "Model_kit.h"
#include "Vector_3D_impl.h"
#include "Form_main_window.h"
#include "Command_template.h"

namespace MM
{

Form_add_model::
Form_add_model ()
{
    setupUi (this);

    toolButton_in->setIcon (QIcon((Path::icon() + "in.png").c_str()));
    connect (toolButton_in,                     SIGNAL(clicked()), 
        this, SLOT(in_clicked()));

    toolButton_right_X->setIcon (QIcon((Path::icon() + "right_X.png").c_str()));
    connect (toolButton_right_X,                SIGNAL(clicked()), 
        this, SLOT(right_X_clicked()));

    toolButton_left_X->setIcon (QIcon((Path::icon() + "left_X.png").c_str()));
    connect (toolButton_left_X,                 SIGNAL(clicked()), 
        this, SLOT(left_X_clicked()));

    toolButton_left_Y->setIcon (QIcon((Path::icon() + "left_Y.png").c_str()));
    connect (toolButton_left_Y,                 SIGNAL(clicked()), 
        this, SLOT(left_Y_clicked()));

    toolButton_left->setIcon (QIcon((Path::icon() + "left.png").c_str()));
    connect (toolButton_left,                   SIGNAL(clicked()), 
        this, SLOT(left_clicked()));

    toolButton_right->setIcon (QIcon((Path::icon() + "right.png").c_str()));
    connect (toolButton_right,                  SIGNAL(clicked()), 
        this, SLOT(right_clicked()));

    toolButton_right_Y->setIcon (QIcon((Path::icon() + "right_Y.png").c_str()));
    connect (toolButton_right_Y,                SIGNAL(clicked()), 
        this, SLOT(right_Y_clicked()));

    toolButton_left_Z->setIcon (QIcon((Path::icon() + "left_Z.png").c_str()));
    connect (toolButton_left_Z,                 SIGNAL(clicked()), 
        this, SLOT(left_Z_clicked()));

    toolButton_out->setIcon (QIcon((Path::icon() + "out.png").c_str()));
    connect (toolButton_out,                    SIGNAL(clicked()), 
        this, SLOT(out_clicked()));

    toolButton_down->setIcon (QIcon((Path::icon() + "down.png").c_str()));
    connect (toolButton_down,                   SIGNAL(clicked()), 
        this, SLOT(down_clicked()));

    toolButton_right_Z->setIcon (QIcon((Path::icon() + "right_Z.png").c_str()));
    connect (toolButton_right_Z,                SIGNAL(clicked()), 
        this, SLOT(right_Z_clicked()));

    connect (pushButton_OK,                     SIGNAL(clicked()), 
        this, SLOT(pushButton_OK_clicked()));

    connect (checkBox_No_overlap,               SIGNAL(toggled(bool)), 
        this, SLOT(checkBox_No_overlap_toggled(bool)));

    connect (lineEdit_Translation_increment,    SIGNAL(textChanged(const QString&)),
        this, SLOT(lineEdit_Translation_increment_textChanged(const QString&)));

    connect (lineEdit_Rotation_increment,       SIGNAL(textChanged(const QString&)), 
        this, SLOT(lineEdit_Rotation_increment_textChanged(const QString&)));

    toolButton_up->setIcon (QIcon((Path::icon() + "up.png").c_str()));
    connect (toolButton_up,                     SIGNAL(clicked()), 
        this, SLOT(up_clicked()));

    toolButton_flush_camera->setIcon (QIcon((Path::icon() + "flush_camera.png").c_str()));
    connect (toolButton_flush_camera,           SIGNAL(clicked()), 
        this, SLOT(flush_camera_clicked()));

    connect (pushButton_Cancel,                 SIGNAL(clicked()), 
        this, SLOT(pushButton_Cancel_clicked()));

//    update_me_ = create_command (*this, &Form_add_model::update);
//    Project::singleton().on_change_current (update_me_);

    lineEdit_Translation_increment->setValidator (new QDoubleValidator (this)); 
    lineEdit_Rotation_increment->   setValidator (new QDoubleValidator (this)); 

//    update ();
//    rebuild ();//fix to show event
}

Form_add_model::
~Form_add_model()
{
//    Project::singleton().on_change_current (update_me_, false);
//    update_me_ = 0;
}

void Form_add_model::
showEvent (QShowEvent * event)
{
    //update_form ();
    if (Project::singleton().model_count() == 0)
        reset_current_model (Model::none());
    else
        reset_current_model (Project::singleton().current_model());

    QWidget::showEvent(event);
}

void Form_add_model::
reset_current_model (Model & model)
{
    if (&model == &Model::none())
        setEnabled (false);
    else
    {
        setEnabled (true);
        update_form ();
    }
}

void Form_add_model::
update_form ()
{
    lineEdit_Translation_increment->setText 
        (QString::number(builder().translation_increment()));
    
    lineEdit_Rotation_increment->setText 
        (QString::number(builder().rotation_increment() * 180. / __Pi__));

    checkBox_No_overlap->setChecked (!builder().is_overlap_allowed());

    if (builder ().is_active())
    {
        //pushButton_OK->      setEnabled (true);
        //pushButton_Cancel->  setEnabled (true);
        //checkBox_No_overlap->setEnabled (true); 
        frame_2-> show ();
    }
    else
    {
        //pushButton_OK->      setEnabled (false);
        //pushButton_Cancel->  setEnabled (false);
        //checkBox_No_overlap->setEnabled (false);
        frame_2-> hide ();
    }
}

void Form_add_model::
pop_up ()
{
    Form_main_window &main_window 
        = Qt_application::singleton().main_window();

    if (Form_add_model::exist())
        main_window.pop_up (Form_add_model::singleton());
    else
        main_window.add_bottom_control
            (Form_add_model::singleton(), "Add model");
    
    //Form_add_model::singleton()->update_form ();
}

Model & Form_add_model::
model ()
{
    return Project::singleton().current_model();
}

Add_builder & Form_add_model::
builder ()
{
    Model & model = Project::singleton().current_model();
    return  model.kit().add_builder();
}

void Form_add_model::up_clicked    ()    {builder ().up   ();}
void Form_add_model::down_clicked  ()    {builder ().down ();}
void Form_add_model::right_clicked ()    {builder ().right();}
void Form_add_model::left_clicked  ()    {builder ().left ();}
void Form_add_model::out_clicked   ()    {builder ().out  ();}
void Form_add_model::in_clicked    ()    {builder ().in   ();}

void Form_add_model::right_X_clicked ()  {builder ().right_X();}
void Form_add_model::left_X_clicked  ()  {builder ().left_X ();}
void Form_add_model::right_Y_clicked ()  {builder ().right_Y();}
void Form_add_model::left_Y_clicked  ()  {builder ().left_Y ();}
void Form_add_model::right_Z_clicked ()  {builder ().right_Z();}
void Form_add_model::left_Z_clicked  ()  {builder ().left_Z ();}

void Form_add_model::
flush_camera_clicked ()
{
    model().kit().view().flush_camera();
    update_form ();
}

void Form_add_model::
pushButton_Cancel_clicked ()
{
    builder ().reject ();
    update_form ();
}

void Form_add_model::
pushButton_OK_clicked()
{
    builder ().accept ();
    update_form ();
}

void Form_add_model::
checkBox_No_overlap_toggled (bool on)
{
    builder().allow_overlap (!on);
}


void Form_add_model::
lineEdit_Translation_increment_textChanged (const QString & new_value)
{
    builder().set_translation_increment (new_value.toDouble());
}

void Form_add_model::
lineEdit_Rotation_increment_textChanged (const QString & new_value)
{
    builder().set_rotation_increment (new_value.toDouble() * __Pi__ / 180.);
}

}//MM












