#include "Aminoacid_builder.h"

#include "Path.h"
#include "Aminoacid_name.h"
#include "Atom.h"
#include "Atom_kit.h"
#include "Model.h"
#include "Model_kit.h"
#include "User.h"
#include "Own.h"
#include "Create.h"
#include "Joint.h"
#include "Branch.h"

namespace MM
{
void Aminoacid_builder::
set_model (Model & model)
{
    model_ = &model;
}

void Aminoacid_builder::
set_joint (Atom * joint)
{
    joint_1_ = joint;
}

void Aminoacid_builder::
set_joint (Atom * joint_1, Atom * joint_2)
{
    REQUIRE ("If joint_2 exist then joint_2 exist ", 
        !(joint_1 == 0 && joint_2 != 0));

    joint_1_ = joint_1;
    joint_2_ = joint_2;
}

void Aminoacid_builder::
//add_aminoacid (const Text & residue_name, double phi, double psi, double omega)
add_aminoacid (const Text & residue_name, double , double , double )
{
    //int i, atom_count;

    if (model_ == 0)
    {
        to_user(). warning ("No model associated with Aminoacid_builder");
        return;
    }

    if (! Aminoacid_name::is_valid (residue_name))
    {
        to_user(). warning (
            Text("\"") + residue_name + "\" is not a valid residuename.");
        return;
    }

    if (model_->atom_count() == 0)
        create_billet ();
    else
        FIX;
    
    if (model_->kit().rigid_body().joint_count() == 0)
    {
        to_user(). warning ("No joints in model.");
        return;
    }

    //Model second_model;
    Own <Model> second_model (prototype <Model>().clone());
    second_model().add (Path::aminoacids_20() + residue_name + ".ent");

    Joint & joint_f0 = model_->kit().rigid_body().joint (0);
    Joint & joint_f1 = model_->kit().rigid_body().joint (1);
    Joint & joint_s0 = second_model().kit().rigid_body().joint (0);
    Joint & joint_s1 = second_model().kit().rigid_body().joint (1);

    joint_f0.chain (joint_s0, 1.36); //1.33 //fix
    joint_s1.chain (joint_f1, 1.36);

    model_->merge_model (&second_model());

    //Branch out_f0 (joint_f0.terminal_atom(), joint_f0.stub_atom());
    //Branch out_f1 (joint_f1.terminal_atom(), joint_f1.stub_atom());
    Branch out_s0 (joint_s0.terminal_atom(), joint_s0.stub_atom());
    Branch out_s1 (joint_s1.terminal_atom(), joint_s1.stub_atom());

    /*for (i=0, atom_count=out_f0.atom_count();  i<atom_count; ++i) 
        out_f0.atom(i).kit().set_alive(false);

    for (i=0, atom_count=out_f1.atom_count();  i<atom_count; ++i) 
        out_f1.atom(i).kit().set_alive(false);//*/

    //for (i=0, atom_count=out_s0.atom_count();  i<atom_count; ++i) 
    //    out_s0.atom(i).kit().set_alive(false);

    //for (i=0, atom_count=out_s1.atom_count();  i<atom_count; ++i) 
    //    out_s1.atom(i).kit().set_alive (false);

    model_->remove_deads ();

    //bound
    //kill ends


    /*if (model_->joint(0).first_atom().element() == "H")
    {
        FIX;
    }
    else//*/
    {
//        Joint joint_1, joint_2;
//        break_bond (model_->joint(0).first_atom(), 
//                    model_->joint(0).second_atom(), 
//                    joint_1, joint_2);
    }

    //fix continue
}

void Aminoacid_builder::
create_billet ()
{
    model_->add (Path::aminoacids_billets() + "AcNHMe.ent");
}

void Aminoacid_builder::
//break_bond (const Atom & atom_1, const Atom & atom_2, 
//            Joint & new_joint_1, Joint & new_joint_2)
break_bond (const Atom & , const Atom & , 
            Joint & , Joint & )
{
  //  FIX;
}

}//MM
