#include "Test.h"

#include "Log.h"

#include "Path.h"
#include "Model.h"
#include "Model_impl.h"

#include "Aminoacid_builder.h"
#include "Atom.h"
#include "Atom_kit.h"
#include "Create.h"
#include "Joint.h"
#include "Float_joint.h"

#include "User.h"

namespace MM
{

class Aminoacid_builder_test : public Test
{
    Aminoacid_builder *link;

public:
    Aminoacid_builder_test( const Text & suite_name ) : Test( suite_name ) { }
    
    void run()
    {
        test_billet();          count_tests();
        test_gly_prototype();   count_tests();
        test_gly();             count_tests();
        test_gly_ala();         count_tests();
    }

    void test_billet();
    void test_gly_prototype();
    void test_gly();
    void test_gly_ala();
};

Aminoacid_builder_test test_aminoacid_builder ("correctness");

void Aminoacid_builder_test::
test_billet()
{
    Own <Model> model (prototype <Model> (). clone ());

    Aminoacid_builder builder;
//    builder.set_model (&(model()));
    builder.set_model (model());
    builder.create_billet ();

    TEST ("AcNHMe was created.",   model().atom_count() == 12);
    TEST ("Model has two joints.", model().kit().rigid_body().joint_count() == 2);
    TEST ("Model has two fixed  joints.", model().kit().rigid_body().fixed_joint_count() == 2);
    TEST ("Model has not float joints.",  model().kit().rigid_body().float_joint_count() == 0);

    //const Atom_group &group = model().joint(0).atom_group();

    /*TEST ("First joint atom 1 is 'C'.", 
        model.joint(0).first_point ().x() == -0.721);

    TEST ("First joint atom 2 is 'N'.", 
        model.joint(0).second_point ().x() == -1.227);

    TEST ("First joint atom 3 is 'CA'.", 
        model.joint(0).third_point ().x() == -1.918);

    TEST ("First joint includes 6 atoms.", 
        model().kit().rigid_body().joint(0).atom_group ().atom_count() == 6);//*/


    
    /*TEST ("Second joint atom 1 is 'N'.", 
        model.joint(1).first_point ().x() == -1.227);

    TEST ("Second joint atom 2 is 'C'.", 
        model.joint(1).second_point ().x() == -0.721);

    TEST ("Second joint atom 3 is 'CH3'.", 
        model.joint(1).third_point ().x() == 0.000);

    TEST ("Second joint includes 7 atoms.", 
        model().kit().rigid_body().joint(1).atom_group ().atom_count() == 6);//*/
}

void Aminoacid_builder_test::
test_gly_prototype ()
{
    Own <Model> model (prototype <Model> (). clone ());
    model().add (Path::aminoacids_20() + "Gly.ent");

    TEST ("Gly was created.",             model().atom_count() == 9);
    TEST ("Model has two joints.",        model().kit().rigid_body().joint_count() == 2);
    TEST ("Model has not fixed  joints.", model().kit().rigid_body().fixed_joint_count() == 0);
    TEST ("Model has two float joints.",  model().kit().rigid_body().float_joint_count() == 2);

    //REMARK @mm joint float 1 2 4 all
    //REMARK @mm joint float 9 7 4 all

    TEST ("First joint atom 1 is 'H'.", 
        model().kit().rigid_body().float_joint(0).first_atom (). kit().pdb().name() == "H");

    TEST ("First joint atom 2 is 'N'.", 
        model().kit().rigid_body().float_joint(0).second_atom ().kit().pdb().name() == "N");

    TEST ("First joint atom 3 is 'CA'.", 
        model().kit().rigid_body().float_joint(0).third_atom (). kit().pdb().name() == "CA");

    //TEST ("First joint includes 9 atoms.", 
    //    model().kit().rigid_body().joint(0).atom_group ().atom_count() == 9);


    
    TEST ("Second joint atom 1 is 'N'.", 
        model().kit().rigid_body().float_joint(1).first_atom (). kit().pdb().name() == "H");

    TEST ("Second joint atom 2 is 'C'.", 
        model().kit().rigid_body().float_joint(1).second_atom ().kit().pdb().name() == "C");

    TEST ("Second joint atom 3 is 'CH3'.", 
        model().kit().rigid_body().float_joint(1).third_atom (). kit().pdb().name() == "CA");

    //TEST ("Second joint includes 9 atoms.", 
    //    model().kit().rigid_body().joint(1).atom_group ().atom_count() == 9);
}
    
void Aminoacid_builder_test::
test_gly()
{
    Own <Model> model (prototype <Model> (). clone ());
    Aminoacid_builder   builder;
    builder.set_model (model());

    TEST ("Empty model.", model().atom_count () == 0);

    builder.add_aminoacid ("Gly");

    //to_user().status(model.atom_count ());
    //TEST ("Billet + Gly in model.", model().atom_count () == 21);
    model().save_as (Path::tests_rubbish() + "tmp.ent");
/*model.save_as( Text(DIR_tests_rubbish) +"Gly_test.ent" );


    TEST("Gly has 7+2 atoms", model.atoms().size() == 7+2 );

    TEST("Gly was created.", gly_was_created( model() ) );//*/
}

void Aminoacid_builder_test::
test_gly_ala()
{
    /*MM_model d_model( new Atom_repository( new Zhorov_atom( Element::Dummy )) );
    Zhorov_model model( d_model );

    Zhorov_amino_acid_builder   model_builder( model );

    model_builder.add_aminoacid( "Gly" );
    model_builder.add_aminoacid( "Gly" );
    model.save_as( Text(DIR_tests_rubbish) +"Gly_Gly_test.ent" );

    TEST("GlyGly has 14+2 atoms", model.atoms().size() == 14+2 );//*/
}


namespace
{
bool gly_was_created(Model &);

//bool gly_was_created(Model & model) 
bool gly_was_created(Model & ) 
{
    /*Collection_of <Atom> & atoms = model.atoms() ;
    for (atoms.first (); atoms.is_valid(); atoms.next() ) 
    {

        Ref<Zhorov_terminal_atom> ref = 
            atoms.current().aspect().zhorov_terminal_atom();

        assert( ref.is_valid() ); 
            
        Zhorov_terminal_atom & zhorov_terminal_atom = ref();
        //zhorov_terminal_atom.name();

    }
//*/
    return true;
}

}//namespace

}//MM
