#include "Form_sequence_editor.h"

#include "Form_sequence_details.h"
#include "Sequence_model.h"

#include "Sequence_editor.h"
#include "Form_main_window.h"
#include "Qt_application.h"
#include "Qt_command.h"
#include "Project.h"
#include "Model.h"
#include "Model_kit.h"
#include "Model_impl.h"
#include "Molecule.h"
#include "Molecular_type.h"
#include "User.h"
#include "Path.h"
#include "Defs.h"

#include "Create.h" //fix remove

#include <qtextedit.h>
#include <QToolButton>
//#include <QAbstractItemModel>
#include <QHeaderView>
#include <QItemDelegate>

#include <QDirModel>

namespace MM
{

Form_sequence_editor::
Form_sequence_editor ()
:
    conformational_type_(stub_model_),
    current_conformation_n_(-1),
    mol_n_(-1),
    res_n_(-1)
{
    setupUi (this);

    button_group_ = new QButtonGroup;
    button_group_->addButton (radioButton_New);
    button_group_->addButton (radioButton_Append);
    button_group_->addButton (radioButton_Prepend);
    button_group_->addButton (radioButton_Mutate);
    button_group_->addButton (radioButton_Branch);

    radioButton_Prepend->setEnabled (false);
    radioButton_Mutate-> setEnabled (false);
    radioButton_Branch-> setEnabled (false);
    pushButton_Delete->  setEnabled (false);
    spinBox->            setEnabled (false);
    //pushButton_Advanced->setEnabled (false);
    radioButton_Branch->  hide ();
    spinBox->             hide ();
    pushButton_Advanced-> hide ();

    //class Sequence_model : public QAbstractItemModel, protected DbC
    sequence_model_ = new Sequence_model;
    tableView_Sequence->setModel (sequence_model_);
    //tableView_Sequence->header()->setResizeMode (QHeaderView::ResizeToContents);
    //tableView_Sequence->horizontalHeader()->setResizeMode (QHeaderView::ResizeToContents);
    //tableView_Sequence->horizontalHeader()->setResizeMode (QHeaderView::Interactive);
    
    QFont font = tableView_Sequence->horizontalHeader()->font();
    font.setPointSize (7);
    tableView_Sequence->horizontalHeader()->setFont (font);
    tableView_Sequence->horizontalHeader()->hide();
    //tableView_Sequence->horizontalHeader()->setDefaultAlignment (Qt::AlignLeft);

    tableView_Sequence->horizontalHeader()->setMinimumSectionSize (4);
    //tableView_Sequence->horizontalHeader()->setDefaultSectionSize (2);
    //tableView_Sequence->verticalHeader()->setMinimumSectionSize (2);
    //tableView_Sequence->verticalHeader()->setDefaultSectionSize (2);
    tableView_Sequence->resizeRowsToContents();
    //tableView_Sequence->setIconSize(QSize(8,8));
    //tableView_Sequence->resizeColumnsToContents();
    tableView_Sequence->setShowGrid (false);
    tableView_Sequence->setGridStyle (Qt::NoPen);

    tableWidget_Current_conformation->verticalHeader  ()->hide();
    tableWidget_Current_conformation->horizontalHeader()->hide();

    tableWidget_Geometry_parameters->verticalHeader  ()->hide();
    tableWidget_Geometry_parameters->horizontalHeader()->hide();

    class Delegate : public QItemDelegate
    {
        QSize sizeHint (const QStyleOptionViewItem & option, 
                        const QModelIndex & index ) const
        {
            QString str = index.data().toString();

            if (str.length() == 1)
            {
                int width = option.fontMetrics.width (str);
                return QSize (width+2, option.fontMetrics.height());
            }
            else
                return QItemDelegate::sizeHint (option, index) - QSize(6, 0);
        }
    };
    tableView_Sequence->setItemDelegate (new Delegate);
    //QAbstractItemDelegate * delegate = tableView_Sequence->itemDelegate ();

    connect (tableView_Sequence, SIGNAL(clicked        (const QModelIndex &)),
             this,               SLOT  (select_residue (const QModelIndex &)));



    connect (listWidget_Class,  SIGNAL(itemDoubleClicked (QListWidgetItem *)),
             this,              SLOT  (change_dir        (QListWidgetItem *)));

    /*connect (
        listWidget_Class,  
        SIGNAL (currentItemChanged (QListWidgetItem *, QListWidgetItem *)),
        this,
        SLOT   (update_residues ())
            );//*/
    connect (listWidget_Class,      SIGNAL (itemClicked (QListWidgetItem *)),
            this,                   SLOT   (update_residues ()));

    connect (pushButton_Advanced,   SIGNAL (clicked ()),
            this,                   SLOT   (show_details ()));



    connect (pushButton_Apply,      SIGNAL (clicked ()),
             this,                  SLOT   (apply ()));
    connect (lineEdit,              SIGNAL (returnPressed ()),
             this,                  SLOT   (apply ()));
    

    //connect (pushButton_New,        SIGNAL (clicked ()),
    //         this,                  SLOT   (new_molecule ()));


    connect (listWidget_Conformational_type,    
             SIGNAL (currentItemChanged (QListWidgetItem*, QListWidgetItem*)),
             this,                  SLOT   (flush_current_conformation ()));

    connect (tableWidget_Current_conformation,  
             SIGNAL (currentCellChanged (int, int, int, int)),
             this,                  SLOT   (change_current_conformation ()));

    radioButton_Append->setChecked(true);
    checkBox_Auto->setCheckState (Qt::Checked);

    repository_dir_ = QDir (Path::repository().c_str());
    update_class ();
}

void Form_sequence_editor::
pop_up ()
{
    Form_main_window &main_window   
        = Qt_application::singleton().main_window();

    if (exist())
        main_window.pop_up (singleton());
    else
        main_window.add_bottom_control (singleton(), "Sequence");
}

void Form_sequence_editor::
showEvent (QShowEvent * event)
{
    if (Project::singleton().model_count() == 0)
        reset_current_model (Model::none());
    else
        reset_current_model (Project::singleton().current_model());

    QWidget::showEvent (event);
}

void Form_sequence_editor::
reset_current_model (Model &)
{

    sequence_model_->reset_all();
    tableView_Sequence->resizeRowsToContents();
    //tableView_Sequence->resizeColumnsToContents();

    int count = tableView_Sequence->horizontalHeader()->count();
    for (int i=0;  i<count;  ++i)
    {
        tableView_Sequence->resizeColumnToContents(i);
    }//*/

    if (mol_n_ >=0 && res_n_ >=0)
        //tableView_Sequence->setCurrentCell (mol_n_, res_n_);
        tableView_Sequence->setCurrentIndex 
            (tableView_Sequence->model()->index(mol_n_, res_n_+1));
}

void Form_sequence_editor::
reset (Model_event::Hint hint)
{
    if (hint == Model_event::Force_Field_changed)
        return;

    if (hint == Model_event::Structure_changed)
        sequence_model_->reset_all();

    else
        update_form ();
}

void Form_sequence_editor::
new_sequence ()
{
/*    using namespace MM;
    Text name;
    to_user().inquiry ("Set chain name.", name);
//    chain_.adopt (new Chain(name, Model_manager::singleton().current())); //fix
    chain_.adopt (new Chain(name, nil<Model>()));
    update ();//*/
}

void Form_sequence_editor::
delete_sequence ()
{
    //update_class ();
}

void Form_sequence_editor::
show_details ()
{
    Form_sequence_details::pop_up ();
}

void Form_sequence_editor::
select_residue (const QModelIndex & index)
{
    mol_n_ = index.row();
    res_n_ = index.column();

    to_user().status (index.column());
}

void Form_sequence_editor::
choose_residue (Text const & name)
{
    Model_impl model;
    model.load (name, "auto detect");

    if (model.kit().has_short_alias())
    {
        lineEdit->insert (model.kit().short_alias().c_str());
    }
    else if (model.kit().has_alias())
    {
        lineEdit->insert (model.kit().alias().c_str());
    }
    else
    {
        Text message;
        message += "Model '";
        message += name;
        message += "' does not have alias.";
        to_user().error (message);
    }

    if (checkBox_Auto->checkState() == Qt::Checked)
        apply ();
}

void Form_sequence_editor::
apply ()
{
    int                     i, index = 0;
    Array_of_own <Model>    model_list;
    //Array_of_ref <Model>    seq_list;
    Array <Text>            seq_list;
    QDir                    dir = residues_directory ();

    QStringList nameFilters;
    nameFilters << "*.mlm";

    QFileInfoList file_list = dir.
        entryInfoList (nameFilters, QDir::Files, QDir::Name);

    // list of resudues prototypes
    int size = file_list.size();
    for (i=0;  i<size;  ++i)
    {
        QFileInfo const & info = file_list.at(i);
        
        //if (info.completeSuffix() == "conformation.mlm")
        //    continue;

        Text file_name = info.absoluteFilePath().ascii();
        model_list.push_back   (new Model_impl);
        model_list.back().load (file_name, "auto detect");
    }

    // list of resudues from the seq-text and prototypes
    QString seq_text = lineEdit->text();
    
    for (i=0;  i<seq_text.size();  )    //next residue
    {
        Text    alias = "";
        bool    res_found = false;

        for (int j=0;  j<model_list.size();  ++j)
        {
            Model_kit const & kit = model_list[j].kit();

            if (comboBox_alias->currentText() == "Short alias" 
                && kit.has_short_alias())

                alias = kit.short_alias();
            else if (kit.has_alias())
                alias = kit.alias();
            else
            {
                Text message;
                message += "Model '";
                message += file_list.at(j).absoluteFilePath().ascii();
                message += "' does not have alias.";
                to_user().error (message);
                return;
            }

            index = seq_text.indexOf (alias.c_str(), i, Qt::CaseInsensitive);

            if (i==index)
            {
                res_found = true;
                //seq_list.push_back (model_list[j]);
                seq_list.push_back(file_list.at(j).absoluteFilePath().ascii());
                i += alias.size();
                break;
            }
        }
        if (!res_found)
        {
            Text message;
            message += "Alias for '";
            message += seq_text.section('\n', i).ascii();
            message += "' was not found.";
            to_user().error (message);
            return;
        }
    }

    // apply list of resudues
    QAbstractButton *checkedButton = button_group_->checkedButton ();
    CHECK("Some button always checked.", checkedButton != 0);
    QString action = checkedButton->text();
    Sequence_editor & editor = current_model_->kit().sequence_editor ();

    // Define current residue
    mol_n_ = editor.molecule_count() - 1;
    res_n_ = mol_n_ >= 0 ? editor.molecule(mol_n_).residue_count() - 1 : -1;

    QModelIndex currentIndex = tableView_Sequence->currentIndex ();
    if (currentIndex.isValid ())
    {
        mol_n_ = currentIndex.row();
        res_n_ = currentIndex.column() - 1;

        if (res_n_ >= editor.molecule(mol_n_).residue_count())
            res_n_ =  editor.molecule(mol_n_).residue_count() - 1;
    }

    // Apply
    for (int i=0;  i<seq_list.size();  ++i)
    {
        if (action == "New" || mol_n_ == -1)
        {
            editor.new_molecule (seq_list[i]);
            ++mol_n_;
            res_n_ = editor.molecule(mol_n_).residue_count() - 1;
        }
        else if (action == "Append")
        {
            //if (res_n == -1)      //fix to insert
            if (true)
                res_n_ = editor.molecule(mol_n_).residue_count() - 1;

            //editor.append (seq_list[i], current_molecule_n_, current_residue_n_);
            int added = editor.append (mol_n_, res_n_, seq_list[i]);
            res_n_ += added;
            //select res_n
        }
        else if (action == "Prepend")
        {
            if (res_n_ == -1)
                res_n_ = 0;

        }
        else if (action == "Mutate")
        {
            CHECK ("Explicit residue.", res_n_ != -1);
        }
        else if (action == "Branch")
        {
            CHECK ("Explicit residue.", res_n_ != -1);
        }
        else 
            FLAW (Text("No such action: ") + action.ascii());

        if (current_conformation_n_ >= 0 && res_n_ > 0) //fix res_n >= 0
        {
            Molecule     & mol          = editor.molecule (mol_n_);
            Torsion_list & torsion_list = conformational_type_.torsion_list();

            torsion_list.apply (mol, res_n_ - 1, current_conformation_n_);

            ++current_conformation_n_;
            if (current_conformation_n_ >= 
                conformational_type_.torsion_list().residue_count())
                current_conformation_n_ = 0;

            update_current_conformation ();
        }
    }

    lineEdit->clear();
    reset_current_model (*current_model_);
    current_model_->kit().view().fit_to_screen();
}

void Form_sequence_editor::
new_molecule ()
{
    Molecule * molecule = prototype <Molecule>(). clone();
    current_model_->add_molecule (molecule);
    
    sequence_model_->reset_all ();
}

void Form_sequence_editor::
update_class ()
{

    //treeView_Sequence->reset();
    tableView_Sequence->reset();

/*    if (chain_.chain_count () == 0)
    {
        textEdit->setText ("No sequence");
        textEdit->setEnabled (false);
    }
    else
        textEdit->setEnabled (true);//*/
    
    QStringList chemical_classes = repository_dir_.
        entryList (QDir::AllDirs, QDir::Name);

    listWidget_Class->clear ();
    listWidget_Class->addItems (chemical_classes);

    update_residues ();
}

QDir Form_sequence_editor::
residues_directory ()
{
    QDir residues_dir = QDir (repository_dir_);
    QListWidgetItem * item = listWidget_Class->currentItem();

    QString dir_name;
    if (item == 0)
        dir_name = ".";
    else
    {
        dir_name = listWidget_Class->currentItem()->text();

        if (dir_name == "..")
            return residues_dir;
    }

    residues_dir.cd (dir_name);
    return residues_dir;
}

void Form_sequence_editor::
update_residues ()
{
    current_conformation_n_ = -1;
    QDir residues_dir = QDir (repository_dir_);
    QListWidgetItem * item = listWidget_Class->currentItem();

    QString dir_name;
    if (item == 0)
        dir_name = ".";
    else
    {
        dir_name = listWidget_Class->currentItem()->text();

        if (dir_name == "..")
            return;
    }

    residues_dir.cd (dir_name);//*/

    //QStringList residues =
    //    residues_dir.entryList (QDir::Files, QDir::Name);

    // Clear

    /*const QObjectList & children = groupBox_Residues->children ();
    for (int i = 0; i < residues_layout_->count(); ++i)
    {
        QLayoutItem * item   = residues_layout_->itemAt(i);
        QWidget     * widget = item->widget();
        residues_layout_->removeItem (item);
        delete item;
        residues_layout_->removeWidget (widget);
        delete widget;

        //delete residues_layout_->itemAt(i)->widget();
        //delete residues_layout_->itemAt(i);
    }//*/
    delete widget_Residues;                                             //fix !!!!!!!!!!!!!!!!!!!!!!!!!!
    widget_Residues = new QWidget;
    frame->layout()->addWidget (widget_Residues);
    residues_layout_ = new QGridLayout (1, 4);
    residues_layout_->setSpacing (5);
    residues_layout_->setMargin  (0);
    widget_Residues->setLayout (residues_layout_);

    QStringList nameFilters;
    nameFilters << "*.demo" << "*.mlm";
    QFileInfoList file_list = residues_dir.
        entryInfoList (nameFilters, QDir::Files, QDir::Name);

    //int size = residues.size();
    int size = file_list.size();
    for (int i=0;  i<size;  ++i)
    {
        //QString const & name = residues.at(i);
        QFileInfo const & info = file_list.at(i);

        QToolButton *button = new QToolButton;
        //button->setText (name);
        button->setText (info.baseName());

        if (info.completeSuffix() == "demo")
            button->setEnabled (false);

//        if (info.completeSuffix() == "conformation.mlm")
//            continue;


        class Choose_residue : public Command 
        {
            Form_sequence_editor &  form_;
            Text                    name_;

        public:
            Choose_residue (Form_sequence_editor &form, Text const& name)
                : form_(form), name_(name) { }

            void execute () {form_.choose_residue (name_);}
        };

        Qt_command *button_command = new Qt_command (button, 
            new Choose_residue (*this, info.absoluteFilePath().ascii()));

        connect (button,            SIGNAL (clicked ()),
                 button_command,    SLOT   (execute ()));

        /*QSize s = button->fontMetrics().size (Qt::TextSingleLine, name);
        s.setWidth (s.width() + 12);
        s.setHeight(button->minimumSize().height());
        button->setMinimumSize (s);//*/
        button->setSizePolicy (QSizePolicy::MinimumExpanding, 
                               QSizePolicy::Minimum);
        //utton->setSizePolicy (QSizePolicy::Expanding, QSizePolicy::Minimum);
        //button->setAutoRaise (true);

        residues_layout_->addWidget (button);
    }

    update_conformational_type ();
}

void Form_sequence_editor::
update_conformational_type ()
{
    QDir dir = residues_directory ();

    QStringList filters;
    filters << "*.conf";
    QFileInfoList file_list = 
        dir.entryInfoList (filters, QDir::Files, QDir::Name);

    /*//fix
    QDirModel *model = new QDirModel (filters, QDir::Files, QDir::Name);
    listView_Conformational_type->setModel (model);//*/

    listWidget_Conformational_type->blockSignals(true);
    listWidget_Conformational_type->clear ();

    int size = file_list.size();
    for (int i=0;  i<size;  ++i)
    {
        QFileInfo const & info = file_list.at(i);

        listWidget_Conformational_type->addItem (info.baseName());
    }//*/
    listWidget_Conformational_type->blockSignals(false);

    update_current_conformation ();
}

void Form_sequence_editor::
flush_current_conformation ()
{
    //if (conformational_type_.torsion_list().residue_count() == 0)
        current_conformation_n_ = -1;
    //else
    //    current_conformation_n_ = 0;

    update_current_conformation ();
}

void Form_sequence_editor::
change_current_conformation ()
{
    current_conformation_n_ = tableWidget_Current_conformation->currentColumn ();
    update_current_conformation ();
}

void Form_sequence_editor::
update_current_conformation ()
{
    QListWidgetItem * item = listWidget_Conformational_type->currentItem();

    tableWidget_Current_conformation-> blockSignals(true);

    stub_model_.clear();

    if (item)
    {
        QString file_name = item->text();
        file_name += ".conf";

        QDir dir = residues_directory ();
        QFileInfo info;
        info.setFile (dir, file_name);

        //conformational_type_.set_name (info.absoluteFilePath().latin1());
        conformational_type_.add (info.absoluteFilePath().latin1());

        Torsion_list & torsion_list = conformational_type_.torsion_list ();
        int items = torsion_list.residue_count();

        if (items == 0)
        {
            current_conformation_n_ = -1;
            tableWidget_Current_conformation->setRowCount    (0);
            tableWidget_Current_conformation->setColumnCount (0);
        }
        else
        {
            if (current_conformation_n_ == -1)
                current_conformation_n_ = 0;
            
            if (current_conformation_n_ >= items)//fix remove
                current_conformation_n_ = 0;
    
            tableWidget_Current_conformation->setColumnCount (items);
            tableWidget_Current_conformation->setRowCount (1);

            for (int i=0;  i<items;  ++i)
            {
                QTableWidgetItem *cell = new QTableWidgetItem(tr("%1").arg(i));
                cell->setTextAlignment(Qt::AlignHCenter | Qt::AlignVCenter);

                tableWidget_Current_conformation->setItem (0, i, cell);
            }

            if (current_conformation_n_ >= 0)
                tableWidget_Current_conformation->setCurrentCell 
                    (0, current_conformation_n_);
        }
    }
    else
        tableWidget_Current_conformation->setRowCount (0);

    tableWidget_Current_conformation-> blockSignals(false);

    tableWidget_Current_conformation->setRowHeight (0, 12);
    //tableWidget_Current_conformation->resizeRowsToContents();
    tableWidget_Current_conformation->resizeColumnsToContents();

    update_geometry_parameters ();
}

void Form_sequence_editor::
update_geometry_parameters ()
{
    int i;
    tableWidget_Geometry_parameters->blockSignals(true);
    tableWidget_Geometry_parameters->setRowCount (0);

    Torsion_list & torsion_list = conformational_type_.torsion_list ();

    if (current_conformation_n_ >= 0)
    {
        tableWidget_Geometry_parameters->setColumnCount (2);
        //tableWidget_Geometry_parameters->setRowCount (residue_count);

        for (i=0, torsion_list.first_torsion (current_conformation_n_);
             torsion_list.is_valid_torsion ();
             torsion_list.next_torsion (), ++i)
        {
            Text   torsion_type;
            double value;
            torsion_list.current_torsion (torsion_type, value);

            QTableWidgetItem *torsion_type_item = 
                new QTableWidgetItem (torsion_type.c_str());
            

            QTableWidgetItem *value_item = 
                new QTableWidgetItem (QString("%1").arg(180*value/__Pi__));

            tableWidget_Geometry_parameters->insertRow (i);
            tableWidget_Geometry_parameters->setItem (i, 0, torsion_type_item);
            tableWidget_Geometry_parameters->setItem (i, 1, value_item);
        }
    }

    tableWidget_Geometry_parameters->resizeColumnsToContents();

    tableWidget_Geometry_parameters->blockSignals(false);
}

void Form_sequence_editor::
change_dir (QListWidgetItem * item)
{
    repository_dir_.cd (item->text());
    update_class ();
}

}//MM
