#ifndef SEQUENCE_MODEL_H
#define SEQUENCE_MODEL_H

#include "DbC.h"
#include "Sequence_editor.h"
#include "Model.h"
#include "Model_kit.h"
#include "Molecule.h"
#include "Project.h"

#include <QAbstractTableModel>

namespace MM
{

class Sequence_model : public QAbstractTableModel, protected DbC
{

        Model & model () const
        {
            Project & project = Project::singleton    ();
            Model   & model   = project.current_model ();

            ENSURE ("", &model != &Model::none());
            return model;
        }

        Sequence_editor & sequence_editor () const
        {
            return model().kit().sequence_editor();
        }

    public:
        void reset_all ()
        {
            reset();
        }

        int rowCount (const QModelIndex &/* parent*/) const  
        {
            return sequence_editor ().molecule_count();
            /*if (parent.isValid())
            {
                if (parent.internalId () == 0)
                {
                    int n =0;
                    return sequence_editor ().molecular_type().
                        type(n).molecule_count();
                }
                else 
                    return 0;
            }
            else
                return sequence_editor ().molecular_type().count();//*/
        }

        int columnCount (const QModelIndex & /*parent*/) const  
        {
            return sequence_editor ().max_sequence()+1;

            /*if (parent.isValid())
            {
                return 5;//fix
                //int n =0;
                //return sequence_editor ().molecular_type().
                //            type(n).molecule_count();
            }
            else
            {
                if (parent.child (0, 0) .isValid())
                    return 10;
                else
                    return 3;
            }//*/
        }

        QVariant data (const QModelIndex & index, int role) const
        {
            if (!index.isValid())
                return QVariant();

            if(role != Qt::DisplayRole)
                return QVariant();
            
            if (index.column() == 0)
            {
                Molecule const& mol = sequence_editor().molecule (index.row());

                //return QString(mol.molecular_formula (Molecule::HTML).c_str());
                return QString(mol.molecular_formula ().c_str());
            }
            else
            {
                Residue const& res = 
                    sequence_editor().residue (index.row(), index.column()-1);

                if (&res == &Residue::none())
                    return QVariant();

                if (res.has_short_alias())
                    return res.short_alias().c_str();
                if (res.has_alias())
                    return QString (res.alias().c_str());
                if (res.has_name())
                    return res.name().c_str();
                return "?";
            }

            /*if (index.internalId () == 0)
            {
                int i=0;
                Molecular_sequence const & seq = 
                    sequence_editor().molecular_type();
                Molecular_type const & type = seq.type (index.row());

                if (index.column() == 0)
                {
                    return QString (type.molecule(0).
                        molecular_formula ().c_str());
                }
                return 123;
            }
            else 
                return index.internalId () - 1;//*/

            return QVariant();
        }
        
        QVariant headerData (int section, 
                             Qt::Orientation orientation, 
                             int role) const
        {
            if (role != Qt::DisplayRole)    
                return QVariant();
                
            if (orientation == Qt::Horizontal)
            {
                switch (section)
                {
                case 0:  return QString ("Mol");
                default: return section;
                }
            }
            else
                return QString("%1").arg(section);
        }

        /*QModelIndex index (int row, int column,
                       const QModelIndex &parent = QModelIndex()) const
        {
            if (!parent.isValid())
                return createIndex (row, column, 0);//column 0

            int i = parent.internalId ();
            if (i != 0)
                return QModelIndex ();//column >1
            //CHECK (Text("i = ") + i + ". Second tree level.", i == 0);

            return createIndex (row, column, row+1);//column 1
        }

        QModelIndex parent (const QModelIndex &index) const
        {
             if (!index.isValid())
                 return QModelIndex();

             int i = index.internalId ();

             if (i == 0)
                 return QModelIndex();//column 0
             else 
                 return createIndex (i-1, 0, 0);//column 1
//                 return createIndex (i-1, index.column(), 0);
//                 return createIndex (0, 0, 0);
        }//*/
};

}//MM

#endif //SEQUENCE_MODEL_H
