#include "Edit_module.h"

#include "Abstracts.h"
#include "Create.h"

#include "Command_template.h"
#include "Bond_detector.h"
#include "Molecular_detector.h"
#include "Project.h"
#include "Model.h"
#include "Model_kit.h"
#include "Fine_type.h"
#include "User.h"

namespace MM
{
Edit_module  Edit_module::singleton_;

Edit_module::Edit_module() : Module ("Edit") { }

void Edit_module::
init ()
{
    adopt_command (prototype <Show_geometry>().clone())
        .set_menu        ("&Edit", "&Geometry")
        //.set_accelerator ("Ctrl+E")
        .set_tool_tip    ("Geometry: distances, angles, dihedrals.")
        .set_icon        ("geometry.png");

    adopt_command (prototype <Show_model_movement>().clone())
        .set_menu        ("&Edit", "&Move model")
        .set_icon        ("out.png");

    adopt_command (prototype <Set_Z_viewpoint>().clone())
        //.set_menu        ("&Edit", "Set Z viewpoint")
        .set_tool_tip    ("Setup viewpoint along Z axis. X - right. Y - up.")
        .set_icon        ("flush_camera.png");

    adopt_command (create_command (*this, &Edit_module::detect_bonds))
        .set_menu        ("&Edit", "Detect bonds");

    adopt_command (create_command (*this, &Edit_module::detect_molecules))
        .set_menu        ("&Edit", "Detect molecules");

    //insert_menu_separator ("&Edit");
}

void Edit_module::
detect_bonds ()
{
    Bond_detector detector (Project::singleton().current_model());
    detector.execute ();
}

void Edit_module::
detect_molecules ()
{
    Molecular_detector detector (Project::singleton().current_model());
    detector.execute ();
}

}//MM








