#ifndef QRMWIDGET_H
#define QRMWIDGET_H

#ifndef DBC_H
#include "DbC.h"
#endif

#include <Qt/qgl.h>
//Added by qt3to4:
#include <QMouseEvent>
#include <QTimerEvent>
#include <QPaintEvent>
#include <QCloseEvent>

extern "C"
{
    struct RMpipe;
    struct RMnode;
    struct RMcamera3D;
}

namespace MM
{
class OpenRM_model_view;
}

class QRMWidget : public QGLWidget, protected MM::DbC
{
    Q_OBJECT

    MM::OpenRM_model_view   *model_view_; //fix to event?

    RMpipe                  *pipe_;

    RMnode                  *root_node_;
    RMnode                  *camera_node_;
    RMnode                  *scene_node_;          
    RMnode                  *object_node_; 
    
    QPoint                  last_position_;

    int                     idle_timer_id_;

public:
    explicit QRMWidget ();
    //explicit QRMWidget (QWidget *parent, const char *name = "");
    ~QRMWidget();

    void            set_model_view (MM::OpenRM_model_view *model_view);
    RMnode *        root_node ()    {return root_node_;}
    RMnode *        scene_node ()   {return scene_node_;}
    RMnode *        object_node ()  {return object_node_;}
    RMpipe *        pipe()          {return pipe_;}

    double          eye_distance (RMcamera3D *camera) const;
    void            set_eye_distance (RMcamera3D *camera, double distance);
    double          eye_distance () const;
    void            set_eye_distance (double distance);

public slots:
    void            fit_center    ();
    void            fit_to_screen ();
    void            flush_camera  ();
    void            demo_rotation (int steps);

    //bool            close ();

protected:
    void            destroy (bool destroyWindow=TRUE, 
                             bool destroySubWindows=TRUE);

    void            paintEvent (QPaintEvent *event) ;

    void            initializeGL ();
    void            resizeGL (int width, int height);
    void            paintGL ();

    //void            keyPressEvent   (QKeyEvent   * event);
    void            mousePressEvent (QMouseEvent * event);
    void            mouseMoveEvent  (QMouseEvent * event);
    void            closeEvent (QCloseEvent * e);
    void            windowActivationChange (bool oldActive);
    void            timerEvent (QTimerEvent *e);

    void            create_lighting ();
    void            create_camera   ();

    bool            invariant () const;
};

#endif //QRMWIDGET_H
