#include "Qt_interface.h"

#include "Qt_command.h"

#include "Module_handle.h"
#include "Qt_application.h"
#include "Form_main_window.h"
#include "Path.h"

#include <QToolBar>
#include <QAction>

namespace MM
{
extern int icon_size;

Command * Qt_interface::
handle_apply (Module_handle & handle)
{
    Qt_command       *command = new Qt_command (handle.command().orphan());
    Form_main_window *main_window = &Qt_application::singleton().main_window();
    
    QAction *qAction = new QAction (main_window->parent());
    QObject::connect (qAction, SIGNAL(triggered()), command, SLOT(execute()));

    if (handle.has_accelerator())
    {
        qAction->setShortcut (QKeySequence (handle.accelerator()));
    }

    if (handle.has_icon())
    {
        QToolBar    *toolbar = get_tool_bar (handle.module_name());

        QIcon new_icon = QIcon((Path::icon() + icon_size + "/" + handle.icon()).c_str());
        new_icon.addFile ((Path::icon() + "none-16.png").c_str());

        //qAction->setIcon (QIcon ((Path::icon() + handle.icon()).c_str()));
        qAction->setIcon (new_icon);
        
        if (handle.has_tool_tip())
            if (handle.has_accelerator())
            {
                Text text = Text(handle.tool_tip()) 
                    + " (" + handle.accelerator() + ")";
                qAction->setToolTip (text.c_str());
            }
            else
                qAction->setToolTip (handle.tool_tip());

        toolbar->addAction (qAction);
    }

    if (handle.has_menu())
    {
        qAction->setText (handle.menu_item());
        main_window->add_to_menu (handle.menu(), qAction);
    }

    return command;
}

void Qt_interface::
handle_insert_menu_separator (Text const& menu)
{
    Qt_application::singleton().main_window().
        insert_menu_separator (menu.c_str());
}

QToolBar * Qt_interface::
get_tool_bar (Text const & title)
{
    Form_main_window *main_window = &Qt_application::singleton().main_window();
    return main_window->tool_bar (title.c_str());
}

}//MM
