#include "Form_build_crystal.h"

#include "Log.h"

#include "Crystal_builder_impl.h"
#include "Qt_application.h"
//#include "Form_build_crystal.h"
#include "Form_main_window.h"
#include "Atom.h"
#include "Atom_kit.h"
#include "Model.h"
#include "Model_kit.h"
#include "Project.h"
#include "Command_template.h"
#include "HSV.h"
#include "View_atom_type_manager.h"
#include "View_atom_type.h"
#include "Mach_eps.h"

#include <qpushbutton.h>
#include <qcombobox.h>
#include <qcheckbox.h>
#include <qlineedit.h>
#include <qvalidator.h>

#include <math.h>

namespace MM
{

Form_build_crystal::
Form_build_crystal ()
{
    setupUi (this);

    pushButton_From_Fractional_to_Cartesian->hide();

    checkBox_Use_PBC->setEnabled (false);
    checkBox_Use_PBC->hide();

    checkBox_Show_border->hide();
    comboBox_border_style->hide();

    textLabel_Crystal_system->hide();
    comboBox_Crystal_system->hide();

    checkBox_same_abc->setChecked (true);
    set_abc_the_same (true);

    connect (lineEdit_beta,                           SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_paramiters(const QString&)));
    connect (lineEdit_alpha,                          SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_paramiters(const QString&)));
    connect (lineEdit_mod_c,                          SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_paramiters(const QString&)));
    connect (lineEdit_mod_b,                          SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_paramiters(const QString&)));
    connect (lineEdit_mod_a,                          SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_paramiters(const QString&)));
    connect (lineEdit_gamma,                          SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_paramiters(const QString&)));
    connect (lineEdit_c_copies,                       SIGNAL(textChanged(const QString&)), this, SLOT(accept_number_of_copies(const QString&)));
    connect (lineEdit_a_copies,                       SIGNAL(textChanged(const QString&)), this, SLOT(accept_number_of_copies(const QString&)));
    connect (lineEdit_b_copies,                       SIGNAL(textChanged(const QString&)), this, SLOT(accept_number_of_copies(const QString&)));
    connect (pushButton_Expand,                       SIGNAL(released()),                  this, SLOT(expand_model()));
    connect (pushButton_From_Fractional_to_Cartesian, SIGNAL(clicked()),                   this, SLOT(from_fractional_to_cartesian()));
    connect (checkBox_same_abc,                       SIGNAL(stateChanged(int)),           this, SLOT(set_abc_the_same(int)));
    connect (checkBox_Preview,                        SIGNAL(stateChanged(int)),           this, SLOT(show_cell_copies(int)));
    connect (comboBox_image_style,                    SIGNAL(activated(int)),              this, SLOT(update_image_style(int)));
    connect (lineEdit_ax,                             SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_vectors(const QString&)));
    connect (lineEdit_ay,                             SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_vectors(const QString&)));
    connect (lineEdit_az,                             SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_vectors(const QString&)));
    connect (lineEdit_bx,                             SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_vectors(const QString&)));
    connect (lineEdit_by,                             SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_vectors(const QString&)));
    connect (lineEdit_bz,                             SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_vectors(const QString&)));
    connect (lineEdit_cx,                             SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_vectors(const QString&)));
    connect (lineEdit_cy,                             SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_vectors(const QString&)));
    connect (lineEdit_cz,                             SIGNAL(textChanged(const QString&)), this, SLOT(accept_lattice_vectors(const QString&)));
    connect (lineEdit_alpha,                          SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_gamma,                          SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_beta,                           SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_az,                             SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_bz,                             SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_cz,                             SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_ay,                             SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_by,                             SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_cy,                             SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_ax,                             SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_bx,                             SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_cx,                             SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_mod_a,                          SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_mod_b,                          SIGNAL(lostFocus()),                 this, SLOT(check_validity()));
    connect (lineEdit_mod_c,                          SIGNAL(lostFocus()),                 this, SLOT(check_validity()));

    class Angles_validator : public QDoubleValidator
    {
        QLineEdit       *alpha_, *beta_, *gamma_;

    public:
        explicit Angles_validator (QLineEdit *alpha, 
                                   QLineEdit *beta, 
                                   QLineEdit *gamma, QObject * parent)
        :   
            QDoubleValidator (parent), 
            alpha_(alpha), beta_(beta), gamma_(gamma) { }

        virtual State validate (QString & input, int & pos) const
        {
            State result = QDoubleValidator::validate (input, pos);
            
            if (result != QValidator::Invalid)
            {
                if (alpha_->text().toDouble() + beta_->text().toDouble() 
                    < gamma_->text().toDouble() ||

                    beta_-> text().toDouble() + gamma_->text().toDouble()
                    < alpha_->text().toDouble() ||

                    alpha_->text().toDouble() + gamma_->text().toDouble()
                    < beta_-> text().toDouble()
                   )
                return QValidator::Intermediate;
            }
            
            return result;
        }
    };
    angles_validator_ = new Angles_validator 
        (lineEdit_alpha, lineEdit_beta, lineEdit_gamma, this);

    comboBox_Crystal_system->insertItem ("Triclinic");
    
    comboBox_border_style  ->insertItem ("No");
    comboBox_border_style  ->insertItem ("Wireframe");
    comboBox_border_style  ->insertItem ("Solid");

    //comboBox_image_style   ->insertItem ("small and pale");
    comboBox_image_style   ->insertItem ("dark");
    comboBox_image_style   ->insertItem ("small and dark");
    comboBox_image_style   ->insertItem ("small");
    comboBox_image_style   ->insertItem ("same");

    //lineEdit_a_copies->setValidator (new QDoubleValidator( 1000, 9999, lineEdit_a_copies ));

    lineEdit_mod_a->setValidator (new QDoubleValidator (1, 1000, 15, this));    
    lineEdit_mod_b->setValidator (new QDoubleValidator (1, 1000, 15, this));     
    lineEdit_mod_c->setValidator (new QDoubleValidator (1, 1000, 15, this));     
    lineEdit_alpha->setValidator (angles_validator_);     
    lineEdit_beta ->setValidator (angles_validator_);     
    lineEdit_gamma->setValidator (angles_validator_);     

    lineEdit_ax->setValidator (new QDoubleValidator (this));     
    lineEdit_ay->setValidator (new QDoubleValidator (this)); 
    lineEdit_az->setValidator (new QDoubleValidator (this)); 
    lineEdit_bx->setValidator (new QDoubleValidator (this)); 
    lineEdit_by->setValidator (new QDoubleValidator (this)); 
    lineEdit_bz->setValidator (new QDoubleValidator (this)); 
    lineEdit_cx->setValidator (new QDoubleValidator (this)); 
    lineEdit_cy->setValidator (new QDoubleValidator (this)); 
    lineEdit_cz->setValidator (new QDoubleValidator (this)); 
    
    lineEdit_a_copies->setValidator (new QIntValidator (1, 100, this));
    lineEdit_b_copies->setValidator (new QIntValidator (1, 100, this));
    lineEdit_c_copies->setValidator (new QIntValidator (1, 100, this));

    //update_form ();

//    update_me_ = create_command (*this, &Form_build_crystal::update);
//    Project::singleton().on_change_current (update_me_);
}

void Form_build_crystal::
showEvent (QShowEvent * event)
{
    //update_form ();
    if (Project::singleton().model_count() == 0)
        reset_current_model (Model::none());
    else
        reset_current_model (Project::singleton().current_model());

    QWidget::showEvent(event);
}

void Form_build_crystal::
check_validity ()
{
    if (
        lineEdit_alpha->hasAcceptableInput () &&
        lineEdit_beta ->hasAcceptableInput () &&
        lineEdit_gamma->hasAcceptableInput ()
       )
       return;

    if (
        lineEdit_mod_a->hasFocus() ||
        lineEdit_mod_b->hasFocus() ||
        lineEdit_mod_c->hasFocus() ||
        lineEdit_alpha->hasFocus() ||
        lineEdit_beta ->hasFocus() ||
        lineEdit_gamma->hasFocus() ||

        lineEdit_ax->hasFocus() ||
        lineEdit_ay->hasFocus() ||
        lineEdit_az->hasFocus() ||
        lineEdit_bx->hasFocus() ||
        lineEdit_by->hasFocus() ||
        lineEdit_bz->hasFocus() ||
        lineEdit_cx->hasFocus() ||
        lineEdit_cy->hasFocus() ||
        lineEdit_cz->hasFocus()
       )
       return;

    to_user(). error 
        ("Angle (alpha or beta or gamma)\n"
         " should be less than sum of two others.");

    lineEdit_alpha->setFocus ();
}

Form_build_crystal::
~Form_build_crystal ()
{
//    Project::singleton().on_change_current (update_me_, false);
//    update_me_ = 0;
}

Crystal_builder & Form_build_crystal::
builder ()
{
    Model & model = Project::singleton().current_model();
    CHECK ("", &model != &Model::none());
    return  model.kit().crystal_builder();
}

void Form_build_crystal::
reset_current_model (Model & model)
{
    if (&model == &Model::none())
        setEnabled (false);
    else
    {
        setEnabled (true);
        update_form ();
    }
}

void Form_build_crystal::
reset (Model_event::Hint hint)
{
    if (hint == Model_event::Structure_changed || 
        hint == Model_event::Conformation_changed)
    {
        builder ().fit_box ();
        update_form ();
        update_model_view ();
    }
}

void Form_build_crystal::
update_form ()
{
    static bool intrance = false;
    if (intrance)                   //fix remove
        return;
    intrance = true;

    /*lineEdit_mod_a->blockSignals (true);
    lineEdit_mod_b->blockSignals (true);
    lineEdit_mod_c->blockSignals (true);
    lineEdit_alpha->blockSignals (true);
    lineEdit_beta ->blockSignals (true);
    lineEdit_gamma->blockSignals (true);

    lineEdit_ax->blockSignals (true);
    lineEdit_ay->blockSignals (true);
    lineEdit_az->blockSignals (true);
    lineEdit_bx->blockSignals (true);
    lineEdit_by->blockSignals (true);
    lineEdit_bz->blockSignals (true);
    lineEdit_cx->blockSignals (true);
    lineEdit_cy->blockSignals (true);
    lineEdit_cz->blockSignals (true);//*/

    /*if (checkBox_same_abc->isChecked ())
        builder().set_number_of_copies 
            (builder().a_copies(),
             builder().a_copies(), 
             builder().a_copies());//*/

    checkBox_Use_PBC    -> setChecked (builder().use_PBC     ());
    checkBox_Show_border-> setChecked (builder().show_border ());
    checkBox_Preview    -> setChecked (builder().show_images ());

    update_lattice_paramiters ();
    //lineEdit_mod_a-> setText (QString::number (builder().a_mod()));
    //lineEdit_mod_b-> setText (QString::number (builder().b_mod()));
    //lineEdit_mod_c-> setText (QString::number (builder().c_mod()));
    //lineEdit_alpha-> setText (QString::number (builder().alpha()));
    //lineEdit_beta -> setText (QString::number (builder().beta ()));
    //lineEdit_gamma-> setText (QString::number (builder().gamma()));

    //lineEdit_mod_a-> setCursorPosition (0);
    //lineEdit_mod_b-> setCursorPosition (0);
    //lineEdit_mod_c-> setCursorPosition (0);
    //lineEdit_alpha-> setCursorPosition (0);
    //lineEdit_beta -> setCursorPosition (0);
    //lineEdit_gamma-> setCursorPosition (0);

    update_lattice_vectors ();
    /*Vector_3D const & a = builder().a ();
    Vector_3D const & b = builder().b ();
    Vector_3D const & c = builder().c ();

    const double zero = ::d_mach_eps_2;
    lineEdit_ax-> setText (a.x() > zero || a.x() < -zero ? QString::number (a.x()) : "0");
    lineEdit_ay-> setText (a.y() > zero || a.y() < -zero ? QString::number (a.y()) : "0");
    lineEdit_az-> setText (a.z() > zero || a.z() < -zero ? QString::number (a.z()) : "0");
    lineEdit_bx-> setText (b.x() > zero || b.x() < -zero ? QString::number (b.x()) : "0");
    lineEdit_by-> setText (b.y() > zero || b.y() < -zero ? QString::number (b.y()) : "0");
    lineEdit_bz-> setText (b.z() > zero || b.z() < -zero ? QString::number (b.z()) : "0");
    lineEdit_cx-> setText (c.x() > zero || c.x() < -zero ? QString::number (c.x()) : "0");
    lineEdit_cy-> setText (c.y() > zero || c.y() < -zero ? QString::number (c.y()) : "0");
    lineEdit_cz-> setText (c.z() > zero || c.z() < -zero ? QString::number (c.z()) : "0");//*/
    //lineEdit_ax-> setCursorPosition (0);
    //lineEdit_ay-> setCursorPosition (0);
    //lineEdit_az-> setCursorPosition (0);
    //lineEdit_bx-> setCursorPosition (0);
    //lineEdit_by-> setCursorPosition (0);
    //lineEdit_bz-> setCursorPosition (0);
    //lineEdit_cx-> setCursorPosition (0);
    //lineEdit_cy-> setCursorPosition (0);
    //lineEdit_cz-> setCursorPosition (0);

    lineEdit_a_copies-> setText (QString::number (builder().a_copies()));
    lineEdit_b_copies-> setText (QString::number (builder().b_copies()));
    lineEdit_c_copies-> setText (QString::number (builder().c_copies()));

    //update_style ();

    /*lineEdit_mod_a->blockSignals (false);
    lineEdit_mod_b->blockSignals (false);
    lineEdit_mod_c->blockSignals (false);
    lineEdit_alpha->blockSignals (false);
    lineEdit_beta ->blockSignals (false);
    lineEdit_gamma->blockSignals (false);

    lineEdit_ax->blockSignals (false);
    lineEdit_ay->blockSignals (false);
    lineEdit_az->blockSignals (false);
    lineEdit_bx->blockSignals (false);
    lineEdit_by->blockSignals (false);
    lineEdit_bz->blockSignals (false);
    lineEdit_cx->blockSignals (false);
    lineEdit_cy->blockSignals (false);
    lineEdit_cz->blockSignals (false);//*/

    intrance = false;               //fix
}

void Form_build_crystal::
update_model_view ()
{
    if (!builder().show_images ())
        return;

    builder().set_show_images (true);//fix

    Array <Model*> const & grid  = builder().grid ();
    Model &          model = builder().model();
    
    double kS = 1, kV = 1, kR = 1;

    if (comboBox_image_style->currentText() == "small and dark")
    {
        //kS = 0.3;
        kS = 1.3;
        kV = .6;
        kR = .5;
    } 
    else if (comboBox_image_style->currentText() == "dark")
    {
        kS = 1.3;
        kV = .6;
    } 
    else if (comboBox_image_style->currentText() == "small")
    {
        kR = .5;
    } 


    for (int n=0;  n<model.atom_count();  ++n)
    {
        Atom &current_atom = model.atom (n);

        View_atom_type type = View_atom_type_manager::instance().
            type (current_atom.element().c_str() );     //fix speed

        double R = type.r(),   G = type.g(),   B = type.b();
        double H,S,V;
        RGB_to_HSV (R,G,B, H,S,V);
        S *= kS;
        V *= kV;
        HSV_to_RGB (H,S,V, R,G,B);
        type.set_color ((int)R,(int)G,(int)B);
        double radius = type.radius();
        type.set_radius (radius * kR);

        for (int i=1;  i<grid.size();  ++i)
        {
            Model & model = *grid [i];
            Atom  & atom  = model.atom (n);
            atom.kit().visual().update_view_type (type);
        }//*/
    }
}

void Form_build_crystal::
update_lattice_vectors ()
{
    lineEdit_ax->blockSignals (true);
    lineEdit_ay->blockSignals (true);
    lineEdit_az->blockSignals (true);
    lineEdit_bx->blockSignals (true);
    lineEdit_by->blockSignals (true);
    lineEdit_bz->blockSignals (true);
    lineEdit_cx->blockSignals (true);
    lineEdit_cy->blockSignals (true);
    lineEdit_cz->blockSignals (true);

    Vector_3D const & a = builder().a ();
    Vector_3D const & b = builder().b ();
    Vector_3D const & c = builder().c ();

    const double zero = ::d_mach_eps_2;
    lineEdit_ax-> setText (a.x() > zero || a.x() < -zero ? QString::number (a.x()) : "0");
    lineEdit_ay-> setText (a.y() > zero || a.y() < -zero ? QString::number (a.y()) : "0");
    lineEdit_az-> setText (a.z() > zero || a.z() < -zero ? QString::number (a.z()) : "0");
    lineEdit_bx-> setText (b.x() > zero || b.x() < -zero ? QString::number (b.x()) : "0");
    lineEdit_by-> setText (b.y() > zero || b.y() < -zero ? QString::number (b.y()) : "0");
    lineEdit_bz-> setText (b.z() > zero || b.z() < -zero ? QString::number (b.z()) : "0");
    lineEdit_cx-> setText (c.x() > zero || c.x() < -zero ? QString::number (c.x()) : "0");
    lineEdit_cy-> setText (c.y() > zero || c.y() < -zero ? QString::number (c.y()) : "0");
    lineEdit_cz-> setText (c.z() > zero || c.z() < -zero ? QString::number (c.z()) : "0");

    lineEdit_ax->blockSignals (false);
    lineEdit_ay->blockSignals (false);
    lineEdit_az->blockSignals (false);
    lineEdit_bx->blockSignals (false);
    lineEdit_by->blockSignals (false);
    lineEdit_bz->blockSignals (false);
    lineEdit_cx->blockSignals (false);
    lineEdit_cy->blockSignals (false);
    lineEdit_cz->blockSignals (false);
}

void Form_build_crystal::
update_lattice_paramiters ()
{
    lineEdit_mod_a->blockSignals (true);
    lineEdit_mod_b->blockSignals (true);
    lineEdit_mod_c->blockSignals (true);
    lineEdit_alpha->blockSignals (true);
    lineEdit_beta ->blockSignals (true);
    lineEdit_gamma->blockSignals (true);

    lineEdit_mod_a-> setText (QString::number (builder().a_mod()));
    lineEdit_mod_b-> setText (QString::number (builder().b_mod()));
    lineEdit_mod_c-> setText (QString::number (builder().c_mod()));
    lineEdit_alpha-> setText (QString::number (builder().alpha()));
    lineEdit_beta -> setText (QString::number (builder().beta ()));
    lineEdit_gamma-> setText (QString::number (builder().gamma()));

    lineEdit_mod_a->blockSignals (false);
    lineEdit_mod_b->blockSignals (false);
    lineEdit_mod_c->blockSignals (false);
    lineEdit_alpha->blockSignals (false);
    lineEdit_beta ->blockSignals (false);
    lineEdit_gamma->blockSignals (false);
}//*/

void Form_build_crystal::
update_image_style (int)
{
    update_model_view ();
}

void Form_build_crystal::
accept_lattice_paramiters (const QString &)     //fix
{
    double mod_a = lineEdit_mod_a->text().toDouble();
    double mod_b = lineEdit_mod_b->text().toDouble();
    double mod_c = lineEdit_mod_c->text().toDouble();
    double alpha = lineEdit_alpha->text().toDouble();
    double beta  = lineEdit_beta ->text().toDouble();
    double gamma = lineEdit_gamma->text().toDouble();

    if (fabs(mod_a) < ::d_mach_eps_2 ||
        fabs(mod_b) < ::d_mach_eps_2 ||
        fabs(mod_c) < ::d_mach_eps_2 ||
        fabs(alpha) < ::d_mach_eps_2 ||
        fabs(beta)  < ::d_mach_eps_2 ||
        fabs(gamma) < ::d_mach_eps_2)
        return;
        
    builder().set (mod_a, mod_b, mod_c, alpha, beta, gamma);
    
    //update_form ();
    update_lattice_vectors ();
}

void Form_build_crystal::
accept_lattice_vectors (const QString & )      //fix
{
    double ax = lineEdit_ax->text().toDouble();
    double ay = lineEdit_ay->text().toDouble();
    double az = lineEdit_az->text().toDouble();

    double bx = lineEdit_bx->text().toDouble();
    double by = lineEdit_by->text().toDouble();
    double bz = lineEdit_bz->text().toDouble();

    double cx = lineEdit_cx->text().toDouble();
    double cy = lineEdit_cy->text().toDouble();
    double cz = lineEdit_cz->text().toDouble();

    if (fabs(ax) < ::d_mach_eps_2 ||
        fabs(by) < ::d_mach_eps_2 ||
        fabs(cz) < ::d_mach_eps_2)
        return;

    builder().set (
        Vector_3D_impl (ax, ay, az),
        Vector_3D_impl (bx, by, bz),
        Vector_3D_impl (cx, cy, cz));

    //update_form ();
    update_lattice_paramiters ();
}

void Form_build_crystal::
accept_number_of_copies (const QString &)         //fix
{

    int new_a_copies = lineEdit_a_copies->text().toInt();
    int new_b_copies = lineEdit_b_copies->text().toInt();
    int new_c_copies = lineEdit_c_copies->text().toInt();

    if (checkBox_same_abc->isChecked ())
        new_c_copies = new_b_copies = new_a_copies;

    if (builder().a_copies() != new_a_copies || 
        builder().b_copies() != new_b_copies ||
        builder().c_copies() != new_c_copies)
    {
        builder().set_number_of_copies 
            (new_a_copies, new_b_copies, new_c_copies);

        update_form ();
        update_model_view ();
    }
}

void Form_build_crystal::
pop_up ()
{
    Form_main_window &main_window 
        = Qt_application::singleton().main_window();

    if (Form_build_crystal::exist())
        main_window.pop_up (Form_build_crystal::singleton());
    else
        main_window.add_left_control
            (Form_build_crystal::singleton(), "Build crystal");
}

void Form_build_crystal::
expand_model ()
{
    pushButton_Expand->setFocus();
    
    int current_item = comboBox_image_style-> currentItem ();
    comboBox_image_style-> setCurrentItem (3);
    update_model_view ();

    builder().expand_model ();
    
    update_form ();
    comboBox_image_style-> setCurrentItem (current_item);
    update_model_view ();
    builder().model().kit().view().fit_to_screen();
}

void Form_build_crystal::
set_abc_the_same (int yes)
{
    if (yes)
    {
        lineEdit_b_copies->setEnabled (false);
        lineEdit_c_copies->setEnabled (false);

        int a_copies = builder().a_copies ();
        builder().set_number_of_copies 
            (a_copies, a_copies, a_copies);
        
        update_model_view ();
    }
    else
    {
        lineEdit_b_copies->setEnabled (true);
        lineEdit_c_copies->setEnabled (true);
    }
    update_form ();
}

void Form_build_crystal::
show_cell_copies (int flag)
{
    builder().set_show_images (flag);
    
    if (flag)
        update_model_view ();
}

void Form_build_crystal::
from_fractional_to_cartesian ()
{
    builder().convert_from_fractional_to_cartesian();
}

}//MM
