#include "M_DynaMix_QProcess.h"

#include "User.h"
#include "Path.h"
#include "Interface_for_M_DynaMix_process.h"
#include "Atom.h"
#include "Atom_group.h"
#include "Text_as_text_output.h"
#include "File_text_input.h"
#include "Collect_potential_from_mDynaMix.h"
#include "Collect_temperature_from_mDynaMix.h"
#include "Collect_pressure_from_mDynaMix.h"
#include "Collect_density_from_mDynaMix.h"
#include "Point_3D_impl.h"

#include <qapplication.h>//fix remove

#include <sstream>

namespace MM
{

M_DynaMix_QProcess::
M_DynaMix_QProcess (Interface_for_M_DynaMix_process & simulator)
:
    M_DynaMix_process   (simulator),
    state_              (none),
    process_            (new Q3Process ()),
    //process_            (0),
    janitor_            (process_),
    ordered_atoms_      (0),//fix check validity
    //output_             (simulator.output_file_name()),
    current_atom_number_(0)
{
}

void M_DynaMix_QProcess::
start (Atom_group * ordered_atoms)
{
    to_user(). status ("Launch M.DynaMix.");
    state_ = none;

    //fix to output_.adopt (new File_text_output (simulator().output_file_name()));
    Text filename;
    filename += Path::m_dynamix();
    filename += simulator().output_file_name();
//    filename += ".tmp";
    output_.adopt (new File_text_output (filename));

    /*if (process_ == 0)
    {
        process_ = new QProcess ();
        janitor_.adopt (process_);
    }//*/

    process_->disconnect ();

    comment_ = Text();

    Text args (Path::m_dynamix());   //fix  make the same in pipe
    #ifdef WIN32
         args += "MDynaMix.exe";
    #else
         args += "md";
    #endif

    //args += " < md.input";

    //3 process_->setArguments (args.c_str());
    process_->clearArguments ();
    process_->addArgument (args.c_str());

    //process_->addArgument ("<");
    //process_->addArgument ("md.input");
    
    connect (process_, SIGNAL (readyReadStdout  ()),    //fix move
             this,     SLOT   (read_from_stdout ()));

    connect (process_, SIGNAL (processExited  ()),
             this,     SLOT   (read_from_stdout ()));

    ordered_atoms_ = ordered_atoms;

    //process_->setCommunication (QProcess::Stdout | QProcess::Stderr);
    //bool success = process_->start();
    //bool success = process_->launch ("");

    Text_as_text_output input;
    File_text_input     file ("MDynaMix.input"); //fix in all places
    const int           bufsize = 255;
    char                buf [bufsize];

    while (file.next_line (buf, bufsize))
        input.text(buf);

    bool success = process_->launch (QString(input.c_str()));
    if (!success)
        to_user().error ("Could not start M.DynaMix.");

    while (process_->isRunning () || process_->canReadLineStdout ())
    {
        qApp->processEvents (QEventLoop::AllEvents, 1);
    }
    for (int i=0;  i<10000;  ++i)
        qApp->processEvents (QEventLoop::AllEvents, 10);
}

/*void M_DynaMix_QProcess::
stop ()
{
}//*/

void M_DynaMix_QProcess::
read_from_stdout()
{
    REQUIRE ("", ordered_atoms_ != 0);

    static bool reinter_flag = false;
    if (reinter_flag)
    {
        //log() << "# reinter #";
        return;
    }
    reinter_flag = true;

    QString result;

    while (process_->canReadLineStdout ())
    {
        //CHECK ("", process_->canReadLineStdout ());

        result = process_->readLineStdout ();

        if (result.isNull())
        {
            to_user() .error ("result.isNull()");
            continue;
        }

        if (result[4] == 'c' && result.find ("coord.end") != -1)
        {
            #ifdef WIN32
//            SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_NORMAL);
            #endif

            state_ = none;
            continue;
        }

        if (state_ == read_coords)
        {
            double x, y, z;
            std::string first_word;
            //std::string line (result.latin1());
            //std::istringstream in (line);
            //std::istringstream in (result);
            std::istringstream in (result.latin1());

            in >> first_word >> x >> y >> z;
            
#ifndef NDEBUG
            //if (in.fail() || first_word != "@mm")
            if (first_word != "@mm")
            {
                to_user (). error (Text () + 
                    "Can not read coords from '" + result.latin1() + "'.");
                simulator().stop();
                break;
            }
#endif
        
            //if (model().atom_count() !=0)
            CHECK ("", ordered_atoms_->atom_count() != 0);
            //if (in.good())
            if (!in.fail())
            {
                Atom &current_atom = ordered_atoms_->atom (current_atom_number_);
                //current_atom.set_x (x);
                //current_atom.set_y (y);
                //current_atom.set_z (z);
                current_atom.set_position (Point_3D_impl (x, y, z));
                //fix to model().set_coords;
            }
        
            ++current_atom_number_;
            continue;
        }

        if (result.find ("@mm") == -1)
        {
            if (result.find ("!!!") != -1)
            {
                if (result.find ("!!! error in input file") != -1)
                {
                    output_().line (result.latin1());
                    comment_ += result.latin1();
                    comment_ += " : ";
            
                    result = process_->readLineStdout ();
                    //output_ << result.latin1() << "\n";
                    output_().line (result.latin1());
                    comment_ += result.latin1();
                    comment_ += " : ";
            
                    result = process_->readLineStdout ();
                    //output_ << result.latin1() << "\n";
                    output_().line (result.latin1());
                    comment_ += result.latin1();
                    comment_ += " : ";
            
                    result = process_->readLineStdout ();
                    comment_ += result.latin1();
                }
                else
                {
                    comment_  = "Warning: ";
                    comment_ += result.latin1();
                }
            }
            
            //output_ << result.latin1() << "\n";
            output_().line (result.latin1());
            output_().flush();
            continue;
        }

        // @mm
        if (result.find ("coord.start") != -1)
        {
            #ifdef WIN32
            ///SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL);
            //SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_HIGHEST);
            #endif

            state_ = read_coords;
            current_atom_number_ = 0;
            continue;
        }

        std::string first_word, second_word;
        std::istringstream line (result.latin1());
        line >> first_word >> second_word;
        
        if (first_word != "@mm")
        {
            to_user (). error (Text () + 
                "No '@mm' prefix in string '" + result.latin1() + "'.");
            simulator().stop();
            break;
        }

        if (second_word == "E")
        {
            Text text;
            double value;

            int number;
            line >> number;   text += "step=";                text += number;
            line >> value;    text += "   E.pot=";        text += value;
            simulator().potential_monitor().set_current_value (value);
            line >> value;    text += "   E.bonded=";   text += value;
            line >> value;    text += "   E.el.st=";      text += value;
            line >> value;    text += "   E.full=";       text += value;
            line >> value;    text += "   T=";            text += value;
            simulator().temperature_monitor().set_current_value (value);
            line >> value;    text += "   P=";            text += value;
            simulator().pressure_monitor().set_current_value (value);
            line >> value;    text += "   d=";            text += value;
            simulator().density_monitor().set_current_value (value);

            /*if (line.fail())
            {
                to_user (). error (Text () + 
                    "Can not E from '" + result.latin1() + "'.");
                simulator().stop();
                break;
            }//*/
            if (line.fail())
            {
                text = result.latin1();
            }

            to_user().status (text);
            simulator().publish_step_finish ();
            //qApp->processEvents (1);

            comment_ = text;
    
            #ifdef WIN32
//            SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_NORMAL);
            //SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_BELOW_NORMAL);
            //SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_LOWEST);
            //SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_IDLE);
            #endif
        }
    }

    reinter_flag = false;
}

}//MM

